@@ -1,6 +1,6 @@
 /*
  * BlueALSA - a2dp-faststream.h
- * Copyright (c) 2016-2021 Arkadiusz Bokowy
+ * Copyright (c) 2016-2024 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -17,13 +17,8 @@
 #endif
 
 #include "a2dp.h"
-#include "ba-transport.h"
 
-extern struct a2dp_codec a2dp_faststream_sink;
-extern struct a2dp_codec a2dp_faststream_source;
-
-void a2dp_faststream_init(void);
-void a2dp_faststream_transport_init(struct ba_transport *t);
-int a2dp_faststream_transport_start(struct ba_transport *t);
+extern struct a2dp_sep a2dp_faststream_source;
+extern struct a2dp_sep a2dp_faststream_sink;
 
 #endif
