@@ -1,6 +1,6 @@
 /*
  * BlueALSA - a2dp-lc3plus.c
- * Copyright (c) 2016-2023 Arkadiusz Bokowy
+ * Copyright (c) 2016-2024 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -9,7 +9,10 @@
  */
 
 #include "a2dp-lc3plus.h"
-/* IWYU pragma: no_include "config.h" */
+
+#if HAVE_CONFIG_H
+# include <config.h>
+#endif
 
 #include <errno.h>
 #include <pthread.h>
@@ -20,13 +23,13 @@
 #include <string.h>
 #include <unistd.h>
 
-#include <glib.h>
-
-#include <lc3.h>
+#include <lc3plus.h>
 
 #include "a2dp.h"
 #include "audio.h"
-#include "bluealsa-config.h"
+#include "ba-config.h"
+#include "ba-transport.h"
+#include "ba-transport-pcm.h"
 #include "io.h"
 #include "rtp.h"
 #include "utils.h"
@@ -36,77 +39,79 @@
 #include "shared/log.h"
 #include "shared/rt.h"
 
-static const struct a2dp_channel_mode a2dp_lc3plus_channels[] = {
-	{ A2DP_CHM_MONO, 1, LC3PLUS_CHANNELS_1 },
-	{ A2DP_CHM_STEREO, 2, LC3PLUS_CHANNELS_2 },
+static const struct a2dp_bit_mapping a2dp_lc3plus_channels[] = {
+	{ LC3PLUS_CHANNEL_MODE_MONO, 1 },
+	{ LC3PLUS_CHANNEL_MODE_STEREO, 2 },
+	{ 0 }
 };
 
-static const struct a2dp_sampling_freq a2dp_lc3plus_samplings[] = {
-	{ 48000, LC3PLUS_SAMPLING_FREQ_48000 },
-	{ 96000, LC3PLUS_SAMPLING_FREQ_96000 },
+static const struct a2dp_bit_mapping a2dp_lc3plus_samplings[] = {
+	{ LC3PLUS_SAMPLING_FREQ_48000, 48000 },
+	{ LC3PLUS_SAMPLING_FREQ_96000, 96000 },
+	{ 0 }
 };
 
-struct a2dp_codec a2dp_lc3plus_sink = {
-	.dir = A2DP_SINK,
-	.codec_id = A2DP_CODEC_VENDOR_LC3PLUS,
-	.capabilities.lc3plus = {
-		.info = A2DP_SET_VENDOR_ID_CODEC_ID(LC3PLUS_VENDOR_ID, LC3PLUS_CODEC_ID),
-		.frame_duration =
-			LC3PLUS_FRAME_DURATION_025 |
-			LC3PLUS_FRAME_DURATION_050 |
-			LC3PLUS_FRAME_DURATION_100,
-		.channels =
-			LC3PLUS_CHANNELS_1 |
-			LC3PLUS_CHANNELS_2,
-		LC3PLUS_INIT_FREQUENCY(
-				LC3PLUS_SAMPLING_FREQ_48000 |
-				LC3PLUS_SAMPLING_FREQ_96000)
-	},
-	.capabilities_size = sizeof(a2dp_lc3plus_t),
-	.channels[0] = a2dp_lc3plus_channels,
-	.channels_size[0] = ARRAYSIZE(a2dp_lc3plus_channels),
-	.samplings[0] = a2dp_lc3plus_samplings,
-	.samplings_size[0] = ARRAYSIZE(a2dp_lc3plus_samplings),
-};
-
-struct a2dp_codec a2dp_lc3plus_source = {
-	.dir = A2DP_SOURCE,
-	.codec_id = A2DP_CODEC_VENDOR_LC3PLUS,
-	.capabilities.lc3plus = {
-		.info = A2DP_SET_VENDOR_ID_CODEC_ID(LC3PLUS_VENDOR_ID, LC3PLUS_CODEC_ID),
-		.frame_duration =
-			LC3PLUS_FRAME_DURATION_025 |
-			LC3PLUS_FRAME_DURATION_050 |
-			LC3PLUS_FRAME_DURATION_100,
-		.channels =
-			LC3PLUS_CHANNELS_1 |
-			LC3PLUS_CHANNELS_2,
-		LC3PLUS_INIT_FREQUENCY(
-				LC3PLUS_SAMPLING_FREQ_48000 |
-				LC3PLUS_SAMPLING_FREQ_96000)
-	},
-	.capabilities_size = sizeof(a2dp_lc3plus_t),
-	.channels[0] = a2dp_lc3plus_channels,
-	.channels_size[0] = ARRAYSIZE(a2dp_lc3plus_channels),
-	.samplings[0] = a2dp_lc3plus_samplings,
-	.samplings_size[0] = ARRAYSIZE(a2dp_lc3plus_samplings),
-};
-
-void a2dp_lc3plus_init(void) {
+static void a2dp_lc3plus_caps_intersect(
+		void *capabilities,
+		const void *mask) {
+	a2dp_caps_bitwise_intersect(capabilities, mask, sizeof(a2dp_lc3plus_t));
 }
 
-void a2dp_lc3plus_transport_init(struct ba_transport *t) {
-
-	const struct a2dp_codec *codec = t->a2dp.codec;
-
-	t->a2dp.pcm.format = BA_TRANSPORT_PCM_FORMAT_S24_4LE;
-	t->a2dp.pcm.channels = a2dp_codec_lookup_channels(codec,
-			t->a2dp.configuration.lc3plus.channels, false);
-	t->a2dp.pcm.sampling = a2dp_codec_lookup_frequency(codec,
-			LC3PLUS_GET_FREQUENCY(t->a2dp.configuration.lc3plus), false);
-
+static int a2dp_lc3plus_caps_foreach_channel_mode(
+		const void *capabilities,
+		enum a2dp_stream stream,
+		a2dp_bit_mapping_foreach_func func,
+		void *userdata) {
+	const a2dp_lc3plus_t *caps = capabilities;
+	if (stream == A2DP_MAIN)
+		return a2dp_bit_mapping_foreach(a2dp_lc3plus_channels, caps->channel_mode, func, userdata);
+	return -1;
 }
 
+static int a2dp_lc3plus_caps_foreach_sampling_freq(
+		const void *capabilities,
+		enum a2dp_stream stream,
+		a2dp_bit_mapping_foreach_func func,
+		void *userdata) {
+	const a2dp_lc3plus_t *caps = capabilities;
+	if (stream == A2DP_MAIN) {
+		const uint16_t sampling_freq = A2DP_LC3PLUS_GET_SAMPLING_FREQ(*caps);
+		return a2dp_bit_mapping_foreach(a2dp_lc3plus_samplings, sampling_freq, func, userdata);
+	}
+	return -1;
+}
+
+static void a2dp_lc3plus_caps_select_channel_mode(
+		void *capabilities,
+		enum a2dp_stream stream,
+		unsigned int channels) {
+	a2dp_lc3plus_t *caps = capabilities;
+	if (stream == A2DP_MAIN)
+		caps->channel_mode = a2dp_bit_mapping_lookup_value(a2dp_lc3plus_channels,
+				caps->channel_mode, channels);
+}
+
+static void a2dp_lc3plus_caps_select_sampling_freq(
+		void *capabilities,
+		enum a2dp_stream stream,
+		unsigned int frequency) {
+	a2dp_lc3plus_t *caps = capabilities;
+	if (stream == A2DP_MAIN) {
+		const uint16_t sampling_freq = a2dp_bit_mapping_lookup_value(a2dp_lc3plus_samplings,
+				A2DP_LC3PLUS_GET_SAMPLING_FREQ(*caps), frequency);
+		A2DP_LC3PLUS_SET_SAMPLING_FREQ(*caps, sampling_freq);
+	}
+}
+
+static struct a2dp_caps_helpers a2dp_lc3plus_caps_helpers = {
+	.intersect = a2dp_lc3plus_caps_intersect,
+	.has_stream = a2dp_caps_has_main_stream_only,
+	.foreach_channel_mode = a2dp_lc3plus_caps_foreach_channel_mode,
+	.foreach_sampling_freq = a2dp_lc3plus_caps_foreach_sampling_freq,
+	.select_channel_mode = a2dp_lc3plus_caps_select_channel_mode,
+	.select_sampling_freq = a2dp_lc3plus_caps_select_sampling_freq,
+};
+
 static bool a2dp_lc3plus_supported(int samplerate, int channels) {
 
 	if (lc3plus_channels_supported(channels) == 0) {
@@ -124,8 +129,9 @@ static bool a2dp_lc3plus_supported(int samplerate, int channels) {
 
 static LC3PLUS_Enc *a2dp_lc3plus_enc_init(int samplerate, int channels) {
 	LC3PLUS_Enc *handle;
+	int32_t lfe_channel_array[1] = { 0 };
 	if ((handle = malloc(lc3plus_enc_get_size(samplerate, channels))) != NULL &&
-			lc3plus_enc_init(handle, samplerate, channels) == LC3PLUS_OK)
+			lc3plus_enc_init(handle, samplerate, channels, 1, lfe_channel_array) == LC3PLUS_OK)
 		return handle;
 	free(handle);
 	return NULL;
@@ -141,7 +147,7 @@ static void a2dp_lc3plus_enc_free(LC3PLUS_Enc *handle) {
 static LC3PLUS_Dec *a2dp_lc3plus_dec_init(int samplerate, int channels) {
 	LC3PLUS_Dec *handle;
 	if ((handle = malloc(lc3plus_dec_get_size(samplerate, channels))) != NULL &&
-			lc3plus_dec_init(handle, samplerate, channels, LC3PLUS_PLC_ADVANCED) == LC3PLUS_OK)
+			lc3plus_dec_init(handle, samplerate, channels, LC3PLUS_PLC_ADVANCED, 1) == LC3PLUS_OK)
 		return handle;
 	free(handle);
 	return NULL;
@@ -167,15 +173,14 @@ static int a2dp_lc3plus_get_frame_dms(const a2dp_lc3plus_t *conf) {
 	}
 }
 
-void *a2dp_lc3plus_enc_thread(struct ba_transport_thread *th) {
+void *a2dp_lc3plus_enc_thread(struct ba_transport_pcm *t_pcm) {
 
 	/* Cancellation should be possible only in the carefully selected place
 	 * in order to prevent memory leaks and resources not being released. */
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
-	pthread_cleanup_push(PTHREAD_CLEANUP(ba_transport_thread_cleanup), th);
+	pthread_cleanup_push(PTHREAD_CLEANUP(ba_transport_pcm_thread_cleanup), t_pcm);
 
-	struct ba_transport *t = th->t;
-	struct ba_transport_pcm *t_pcm = th->pcm;
+	struct ba_transport *t = t_pcm->t;
 	struct io_poll io = { .timeout = -1 };
 
 	const a2dp_lc3plus_t *configuration = &t->a2dp.configuration.lc3plus;
@@ -198,14 +203,10 @@ void *a2dp_lc3plus_enc_thread(struct ba_transport_thread *th) {
 
 	pthread_cleanup_push(PTHREAD_CLEANUP(a2dp_lc3plus_enc_free), handle);
 
-	if ((err = lc3plus_enc_set_frame_ms(handle, lc3plus_frame_dms * 0.1)) != LC3PLUS_OK) {
+	if ((err = lc3plus_enc_set_frame_dms(handle, lc3plus_frame_dms)) != LC3PLUS_OK) {
 		error("Couldn't set frame length: %s", lc3plus_strerror(err));
 		goto fail_setup;
 	}
-	if ((err = lc3plus_enc_set_hrmode(handle, 1)) != LC3PLUS_OK) {
-		error("Couldn't set hi-resolution mode: %s", lc3plus_strerror(err));
-		goto fail_setup;
-	}
 	if ((err = lc3plus_enc_set_bitrate(handle, config.lc3plus_bitrate)) != LC3PLUS_OK) {
 		error("Couldn't set bitrate: %s", lc3plus_strerror(err));
 		goto fail_setup;
@@ -256,18 +257,18 @@ void *a2dp_lc3plus_enc_thread(struct ba_transport_thread *th) {
 	/* RTP clock frequency equal to the RTP clock rate */
 	rtp_state_init(&rtp, samplerate, rtp_ts_clockrate);
 
-	debug_transport_thread_loop(th, "START");
-	for (ba_transport_thread_state_set_running(th);;) {
+	debug_transport_pcm_thread_loop(t_pcm, "START");
+	for (ba_transport_pcm_state_set_running(t_pcm);;) {
 
-		ssize_t samples = ffb_len_in(&pcm);
-		switch (samples = io_poll_and_read_pcm(&io, t_pcm, pcm.tail, samples)) {
+		switch (io_poll_and_read_pcm(&io, t_pcm, &pcm)) {
 		case -1:
 			if (errno == ESTALE) {
 				int encoded = 0;
+				void *scratch = NULL;
 				memset(pcm_ch1, 0, lc3plus_ch_samples * sizeof(*pcm_ch1));
 				memset(pcm_ch2, 0, lc3plus_ch_samples * sizeof(*pcm_ch2));
 				/* flush encoder internal buffers by feeding it with silence */
-				lc3plus_enc24(handle, pcm_ch_buffers, rtp_payload, &encoded);
+				lc3plus_enc24(handle, pcm_ch_buffers, rtp_payload, &encoded, scratch);
 				ffb_rewind(&pcm);
 				continue;
 			}
@@ -278,14 +279,11 @@ void *a2dp_lc3plus_enc_thread(struct ba_transport_thread *th) {
 			continue;
 		}
 
-		ffb_seek(&pcm, samples);
-		samples = ffb_len_out(&pcm);
-
 		/* anchor for RTP payload */
 		bt.tail = rtp_payload;
 
 		const int32_t *input = pcm.data;
-		size_t input_samples = samples;
+		size_t input_samples = ffb_len_out(&pcm);
 		size_t output_len = ffb_len_in(&bt);
 		size_t pcm_frames = 0;
 		size_t lc3plus_frames = 0;
@@ -299,8 +297,9 @@ void *a2dp_lc3plus_enc_thread(struct ba_transport_thread *th) {
 				lc3plus_frames < ((1 << 4) - 1)) {
 
 			int encoded = 0;
-			audio_deinterleave_s24_4le(input, lc3plus_ch_samples, channels, pcm_ch1, pcm_ch2);
-			if ((err = lc3plus_enc24(handle, pcm_ch_buffers, bt.tail, &encoded)) != LC3PLUS_OK) {
+			void *scratch = NULL;
+			audio_deinterleave_s24_4le(pcm_ch_buffers, input, channels, lc3plus_ch_samples);
+			if ((err = lc3plus_enc24(handle, pcm_ch_buffers, bt.tail, &encoded, scratch)) != LC3PLUS_OK) {
 				error("LC3plus encoding error: %s", lc3plus_strerror(err));
 				break;
 			}
@@ -341,7 +340,7 @@ void *a2dp_lc3plus_enc_thread(struct ba_transport_thread *th) {
 				ffb_seek(&bt, rtp_headers_len + chunk_len);
 
 				ssize_t len = ffb_blen_out(&bt);
-				if ((len = io_bt_write(th, bt.data, len)) <= 0) {
+				if ((len = io_bt_write(t_pcm, bt.data, len)) <= 0) {
 					if (len == -1)
 						error("BT write error: %s", strerror(errno));
 					goto fail;
@@ -376,14 +375,14 @@ void *a2dp_lc3plus_enc_thread(struct ba_transport_thread *th) {
 			 * have to append new data to the existing one. Since we do not use
 			 * ring buffer, we will simply move unprocessed data to the front
 			 * of our linear buffer. */
-			ffb_shift(&pcm, samples - input_samples);
+			ffb_shift(&pcm, pcm_frames * channels);
 
 		}
 
 	}
 
 fail:
-	debug_transport_thread_loop(th, "EXIT");
+	debug_transport_pcm_thread_loop(t_pcm, "EXIT");
 fail_ffb:
 	pthread_cleanup_pop(1);
 	pthread_cleanup_pop(1);
@@ -397,15 +396,14 @@ fail_init:
 }
 
 __attribute__ ((weak))
-void *a2dp_lc3plus_dec_thread(struct ba_transport_thread *th) {
+void *a2dp_lc3plus_dec_thread(struct ba_transport_pcm *t_pcm) {
 
 	/* Cancellation should be possible only in the carefully selected place
 	 * in order to prevent memory leaks and resources not being released. */
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
-	pthread_cleanup_push(PTHREAD_CLEANUP(ba_transport_thread_cleanup), th);
+	pthread_cleanup_push(PTHREAD_CLEANUP(ba_transport_pcm_thread_cleanup), t_pcm);
 
-	struct ba_transport *t = th->t;
-	struct ba_transport_pcm *t_pcm = th->pcm;
+	struct ba_transport *t = t_pcm->t;
 	struct io_poll io = { .timeout = -1 };
 
 	const a2dp_lc3plus_t *configuration = &t->a2dp.configuration.lc3plus;
@@ -428,14 +426,10 @@ void *a2dp_lc3plus_dec_thread(struct ba_transport_thread *th) {
 	pthread_cleanup_push(PTHREAD_CLEANUP(a2dp_lc3plus_dec_free), handle);
 
 	const int frame_dms = a2dp_lc3plus_get_frame_dms(configuration);
-	if ((err = lc3plus_dec_set_frame_ms(handle, frame_dms * 0.1)) != LC3PLUS_OK) {
+	if ((err = lc3plus_dec_set_frame_dms(handle, frame_dms)) != LC3PLUS_OK) {
 		error("Couldn't set frame length: %s", lc3plus_strerror(err));
 		goto fail_setup;
 	}
-	if ((err = lc3plus_dec_set_hrmode(handle, 1)) != LC3PLUS_OK) {
-		error("Couldn't set hi-resolution mode: %s", lc3plus_strerror(err));
-		goto fail_setup;
-	}
 
 	ffb_t bt = { 0 };
 	ffb_t bt_payload = { 0 };
@@ -469,11 +463,12 @@ void *a2dp_lc3plus_dec_thread(struct ba_transport_thread *th) {
 	 * not fragmented one or the first fragment of fragmented packet. */
 	bool rtp_media_fragment_skip = false;
 
-	debug_transport_thread_loop(th, "START");
-	for (ba_transport_thread_state_set_running(th);;) {
+	debug_transport_pcm_thread_loop(t_pcm, "START");
+	for (ba_transport_pcm_state_set_running(t_pcm);;) {
 
-		ssize_t len = ffb_blen_in(&bt);
-		if ((len = io_poll_and_read_bt(&io, th, bt.data, len)) <= 0) {
+		ssize_t len;
+		ffb_rewind(&bt);
+		if ((len = io_poll_and_read_bt(&io, t_pcm, &bt)) <= 0) {
 			if (len == -1)
 				error("BT poll and read error: %s", strerror(errno));
 			goto fail;
@@ -512,15 +507,17 @@ void *a2dp_lc3plus_dec_thread(struct ba_transport_thread *th) {
 
 		while (missing_pcm_frames > 0) {
 
-			lc3plus_dec24(handle, bt_payload.data, 0, pcm_ch_buffers, 1);
-			audio_interleave_s24_4le(pcm_ch1, pcm_ch2, lc3plus_ch_samples, channels, pcm.data);
+			void *scratch = NULL;
+			lc3plus_dec24(handle, bt_payload.data, 0, pcm_ch_buffers, scratch, 1);
+			audio_interleave_s24_4le(pcm.data, (const int32_t **)pcm_ch_buffers,
+					channels, lc3plus_ch_samples);
 
 			warn("Missing LC3plus data, loss concealment applied");
 
 			const size_t samples = lc3plus_frame_samples;
 			io_pcm_scale(t_pcm, pcm.data, samples);
 			if (io_pcm_write(t_pcm, pcm.data, samples) == -1)
-				error("FIFO write error: %s", strerror(errno));
+				error("PCM write error: %s", strerror(errno));
 
 			missing_pcm_frames -= lc3plus_ch_samples;
 
@@ -566,8 +563,10 @@ void *a2dp_lc3plus_dec_thread(struct ba_transport_thread *th) {
 		/* Decode retrieved LC3plus frames. */
 		while (lc3plus_frames--) {
 
-			err = lc3plus_dec24(handle, lc3plus_payload, lc3plus_frame_len, pcm_ch_buffers, 0);
-			audio_interleave_s24_4le(pcm_ch1, pcm_ch2, lc3plus_ch_samples, channels, pcm.data);
+			void *scratch = NULL;
+			err = lc3plus_dec24(handle, lc3plus_payload, lc3plus_frame_len, pcm_ch_buffers, scratch, 0);
+			audio_interleave_s24_4le(pcm.data, (const int32_t **)pcm_ch_buffers,
+					channels, lc3plus_ch_samples);
 
 			if (err == LC3PLUS_DECODE_ERROR)
 				warn("Corrupted LC3plus data, loss concealment applied");
@@ -581,7 +580,7 @@ void *a2dp_lc3plus_dec_thread(struct ba_transport_thread *th) {
 			const size_t samples = lc3plus_frame_samples;
 			io_pcm_scale(t_pcm, pcm.data, samples);
 			if (io_pcm_write(t_pcm, pcm.data, samples) == -1)
-				error("FIFO write error: %s", strerror(errno));
+				error("PCM write error: %s", strerror(errno));
 
 			/* update local state with decoded PCM frames */
 			rtp_state_update(&rtp, lc3plus_ch_samples);
@@ -594,7 +593,7 @@ void *a2dp_lc3plus_dec_thread(struct ba_transport_thread *th) {
 	}
 
 fail:
-	debug_transport_thread_loop(th, "EXIT");
+	debug_transport_pcm_thread_loop(t_pcm, "EXIT");
 fail_ffb:
 	pthread_cleanup_pop(1);
 	pthread_cleanup_pop(1);
@@ -608,14 +607,168 @@ fail_init:
 	return NULL;
 }
 
-int a2dp_lc3plus_transport_start(struct ba_transport *t) {
+static int a2dp_lc3plus_configuration_select(
+		const struct a2dp_sep *sep,
+		void *capabilities) {
 
-	if (t->profile & BA_TRANSPORT_PROFILE_A2DP_SOURCE)
-		return ba_transport_thread_create(&t->thread_enc, a2dp_lc3plus_enc_thread, "ba-a2dp-lc3p", true);
+	a2dp_lc3plus_t *caps = capabilities;
+	const a2dp_lc3plus_t saved = *caps;
 
-	if (t->profile & BA_TRANSPORT_PROFILE_A2DP_SINK)
-		return ba_transport_thread_create(&t->thread_dec, a2dp_lc3plus_dec_thread, "ba-a2dp-lc3p", true);
+	/* Narrow capabilities to values supported by BlueALSA. */
+	a2dp_lc3plus_caps_intersect(caps, &sep->config.capabilities);
 
-	g_assert_not_reached();
-	return -1;
+	if (caps->frame_duration & LC3PLUS_FRAME_DURATION_100)
+		caps->frame_duration = LC3PLUS_FRAME_DURATION_100;
+	else if (caps->frame_duration & LC3PLUS_FRAME_DURATION_050)
+		caps->frame_duration = LC3PLUS_FRAME_DURATION_050;
+	else if (caps->frame_duration & LC3PLUS_FRAME_DURATION_025)
+		caps->frame_duration = LC3PLUS_FRAME_DURATION_025;
+	else {
+		error("LC3plus: No supported frame durations: %#x", saved.frame_duration);
+		return errno = ENOTSUP, -1;
+	}
+
+	unsigned int channel_mode = 0;
+	if (a2dp_lc3plus_caps_foreach_channel_mode(caps, A2DP_MAIN,
+				a2dp_bit_mapping_foreach_get_best_channel_mode, &channel_mode) != -1)
+		caps->channel_mode = channel_mode;
+	else {
+		error("LC3plus: No supported channel modes: %#x", saved.channel_mode);
+		return errno = ENOTSUP, -1;
+	}
+
+	unsigned int sampling_freq = 0;
+	if (a2dp_lc3plus_caps_foreach_sampling_freq(caps, A2DP_MAIN,
+				a2dp_bit_mapping_foreach_get_best_sampling_freq, &sampling_freq) != -1)
+		A2DP_LC3PLUS_SET_SAMPLING_FREQ(*caps, sampling_freq);
+	else {
+		error("LC3plus: No supported sampling frequencies: %#x", A2DP_LC3PLUS_GET_SAMPLING_FREQ(saved));
+		return errno = ENOTSUP, -1;
+	}
+
+	return 0;
 }
+
+static int a2dp_lc3plus_configuration_check(
+		const struct a2dp_sep *sep,
+		const void *configuration) {
+
+	const a2dp_lc3plus_t *conf = configuration;
+	a2dp_lc3plus_t conf_v = *conf;
+
+	/* Validate configuration against BlueALSA capabilities. */
+	a2dp_lc3plus_caps_intersect(&conf_v, &sep->config.capabilities);
+
+	switch (conf_v.frame_duration) {
+	case LC3PLUS_FRAME_DURATION_025:
+	case LC3PLUS_FRAME_DURATION_050:
+	case LC3PLUS_FRAME_DURATION_100:
+		break;
+	default:
+		debug("LC3plus: Invalid frame duration: %#x", conf->frame_duration);
+		return A2DP_CHECK_ERR_FRAME_DURATION;
+	}
+
+	if (a2dp_bit_mapping_lookup(a2dp_lc3plus_channels, conf_v.channel_mode) == 0) {
+		debug("LC3plus: Invalid channel mode: %#x", conf->channel_mode);
+		return A2DP_CHECK_ERR_CHANNEL_MODE;
+	}
+
+	uint16_t conf_sampling_freq = A2DP_LC3PLUS_GET_SAMPLING_FREQ(conf_v);
+	if (a2dp_bit_mapping_lookup(a2dp_lc3plus_samplings, conf_sampling_freq) == 0) {
+		debug("LC3plus: Invalid sampling frequency: %#x", A2DP_LC3PLUS_GET_SAMPLING_FREQ(*conf));
+		return A2DP_CHECK_ERR_SAMPLING;
+	}
+
+	return A2DP_CHECK_OK;
+}
+
+static int a2dp_lc3plus_transport_init(struct ba_transport *t) {
+
+	unsigned int channels;
+	if ((channels = a2dp_bit_mapping_lookup(a2dp_lc3plus_channels,
+					t->a2dp.configuration.lc3plus.channel_mode)) == 0)
+		return -1;
+
+	unsigned int sampling;
+	if ((sampling = a2dp_bit_mapping_lookup(a2dp_lc3plus_samplings,
+					A2DP_LC3PLUS_GET_SAMPLING_FREQ(t->a2dp.configuration.lc3plus))) == 0)
+		return -1;
+
+	t->a2dp.pcm.format = BA_TRANSPORT_PCM_FORMAT_S24_4LE;
+	t->a2dp.pcm.channels = channels;
+	t->a2dp.pcm.sampling = sampling;
+
+	return 0;
+}
+
+static int a2dp_lc3plus_source_init(struct a2dp_sep *sep) {
+	if (config.a2dp.force_mono)
+		sep->config.capabilities.lc3plus.channel_mode = LC3PLUS_CHANNEL_MODE_MONO;
+	if (config.a2dp.force_44100)
+		warn("LC3plus: 44.1 kHz sampling frequency not supported");
+	return 0;
+}
+
+static int a2dp_lc3plus_source_transport_start(struct ba_transport *t) {
+	return ba_transport_pcm_start(&t->a2dp.pcm, a2dp_lc3plus_enc_thread, "ba-a2dp-lc3p");
+}
+
+struct a2dp_sep a2dp_lc3plus_source = {
+	.name = "A2DP Source (LC3plus)",
+	.config = {
+		.type = A2DP_SOURCE,
+		.codec_id = A2DP_CODEC_VENDOR_ID(LC3PLUS_VENDOR_ID, LC3PLUS_CODEC_ID),
+		.caps_size = sizeof(a2dp_lc3plus_t),
+		.capabilities.lc3plus = {
+			.info = A2DP_VENDOR_INFO_INIT(LC3PLUS_VENDOR_ID, LC3PLUS_CODEC_ID),
+			.frame_duration =
+				LC3PLUS_FRAME_DURATION_025 |
+				LC3PLUS_FRAME_DURATION_050 |
+				LC3PLUS_FRAME_DURATION_100,
+			.channel_mode =
+				LC3PLUS_CHANNEL_MODE_MONO |
+				LC3PLUS_CHANNEL_MODE_STEREO,
+			A2DP_LC3PLUS_INIT_SAMPLING_FREQ(
+					LC3PLUS_SAMPLING_FREQ_48000 |
+					LC3PLUS_SAMPLING_FREQ_96000)
+		},
+	},
+	.init = a2dp_lc3plus_source_init,
+	.configuration_select = a2dp_lc3plus_configuration_select,
+	.configuration_check = a2dp_lc3plus_configuration_check,
+	.transport_init = a2dp_lc3plus_transport_init,
+	.transport_start = a2dp_lc3plus_source_transport_start,
+	.caps_helpers = &a2dp_lc3plus_caps_helpers,
+};
+
+static int a2dp_lc3plus_sink_transport_start(struct ba_transport *t) {
+	return ba_transport_pcm_start(&t->a2dp.pcm, a2dp_lc3plus_dec_thread, "ba-a2dp-lc3p");
+}
+
+struct a2dp_sep a2dp_lc3plus_sink = {
+	.name = "A2DP Sink (LC3plus)",
+	.config = {
+		.type = A2DP_SINK,
+		.codec_id = A2DP_CODEC_VENDOR_ID(LC3PLUS_VENDOR_ID, LC3PLUS_CODEC_ID),
+		.caps_size = sizeof(a2dp_lc3plus_t),
+		.capabilities.lc3plus = {
+			.info = A2DP_VENDOR_INFO_INIT(LC3PLUS_VENDOR_ID, LC3PLUS_CODEC_ID),
+			.frame_duration =
+				LC3PLUS_FRAME_DURATION_025 |
+				LC3PLUS_FRAME_DURATION_050 |
+				LC3PLUS_FRAME_DURATION_100,
+			.channel_mode =
+				LC3PLUS_CHANNEL_MODE_MONO |
+				LC3PLUS_CHANNEL_MODE_STEREO,
+			A2DP_LC3PLUS_INIT_SAMPLING_FREQ(
+					LC3PLUS_SAMPLING_FREQ_48000 |
+					LC3PLUS_SAMPLING_FREQ_96000)
+		},
+	},
+	.configuration_select = a2dp_lc3plus_configuration_select,
+	.configuration_check = a2dp_lc3plus_configuration_check,
+	.transport_init = a2dp_lc3plus_transport_init,
+	.transport_start = a2dp_lc3plus_sink_transport_start,
+	.caps_helpers = &a2dp_lc3plus_caps_helpers,
+};
