@@ -1,6 +1,6 @@
 /*
  * BlueALSA - a2dp-ldac.c
- * Copyright (c) 2016-2023 Arkadiusz Bokowy
+ * Copyright (c) 2016-2024 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -20,13 +20,13 @@
 #include <sys/ioctl.h>
 #include <unistd.h>
 
-#include <glib.h>
-
 #include <ldacBT.h>
 #include <ldacBT_abr.h>
 
 #include "a2dp.h"
-#include "bluealsa-config.h"
+#include "ba-transport.h"
+#include "ba-transport-pcm.h"
+#include "ba-config.h"
 #include "io.h"
 #include "rtp.h"
 #include "utils.h"
@@ -36,92 +36,86 @@
 #include "shared/log.h"
 #include "shared/rt.h"
 
-static const struct a2dp_channel_mode a2dp_ldac_channels[] = {
-	{ A2DP_CHM_MONO, 1, LDAC_CHANNEL_MODE_MONO },
-	{ A2DP_CHM_DUAL_CHANNEL, 2, LDAC_CHANNEL_MODE_DUAL },
-	{ A2DP_CHM_STEREO, 2, LDAC_CHANNEL_MODE_STEREO },
+static const struct a2dp_bit_mapping a2dp_ldac_channels[] = {
+	{ LDAC_CHANNEL_MODE_MONO, 1 },
+	{ LDAC_CHANNEL_MODE_DUAL, 2 },
+	{ LDAC_CHANNEL_MODE_STEREO, 2 },
+	{ 0 }
 };
 
-static const struct a2dp_sampling_freq a2dp_ldac_samplings[] = {
-	{ 44100, LDAC_SAMPLING_FREQ_44100 },
-	{ 48000, LDAC_SAMPLING_FREQ_48000 },
-	{ 88200, LDAC_SAMPLING_FREQ_88200 },
-	{ 96000, LDAC_SAMPLING_FREQ_96000 },
+static const struct a2dp_bit_mapping a2dp_ldac_samplings[] = {
+	{ LDAC_SAMPLING_FREQ_44100, 44100 },
+	{ LDAC_SAMPLING_FREQ_48000, 48000 },
+	{ LDAC_SAMPLING_FREQ_88200, 88200 },
+	{ LDAC_SAMPLING_FREQ_96000, 96000 },
+	{ LDAC_SAMPLING_FREQ_176400, 176400 },
+	{ LDAC_SAMPLING_FREQ_192000, 192000 },
+	{ 0 }
 };
 
-struct a2dp_codec a2dp_ldac_sink = {
-	.dir = A2DP_SINK,
-	.codec_id = A2DP_CODEC_VENDOR_LDAC,
-	.capabilities.ldac = {
-		.info = A2DP_SET_VENDOR_ID_CODEC_ID(LDAC_VENDOR_ID, LDAC_CODEC_ID),
-		.channel_mode =
-			LDAC_CHANNEL_MODE_MONO |
-			LDAC_CHANNEL_MODE_DUAL |
-			LDAC_CHANNEL_MODE_STEREO,
-		/* NOTE: Used LDAC library does not support
-		 *       frequencies higher than 96 kHz. */
-		.frequency =
-			LDAC_SAMPLING_FREQ_44100 |
-			LDAC_SAMPLING_FREQ_48000 |
-			LDAC_SAMPLING_FREQ_88200 |
-			LDAC_SAMPLING_FREQ_96000,
-	},
-	.capabilities_size = sizeof(a2dp_ldac_t),
-	.channels[0] = a2dp_ldac_channels,
-	.channels_size[0] = ARRAYSIZE(a2dp_ldac_channels),
-	.samplings[0] = a2dp_ldac_samplings,
-	.samplings_size[0] = ARRAYSIZE(a2dp_ldac_samplings),
-};
-
-struct a2dp_codec a2dp_ldac_source = {
-	.dir = A2DP_SOURCE,
-	.codec_id = A2DP_CODEC_VENDOR_LDAC,
-	.capabilities.ldac = {
-		.info = A2DP_SET_VENDOR_ID_CODEC_ID(LDAC_VENDOR_ID, LDAC_CODEC_ID),
-		.channel_mode =
-			LDAC_CHANNEL_MODE_MONO |
-			LDAC_CHANNEL_MODE_DUAL |
-			LDAC_CHANNEL_MODE_STEREO,
-		/* NOTE: Used LDAC library does not support
-		 *       frequencies higher than 96 kHz. */
-		.frequency =
-			LDAC_SAMPLING_FREQ_44100 |
-			LDAC_SAMPLING_FREQ_48000 |
-			LDAC_SAMPLING_FREQ_88200 |
-			LDAC_SAMPLING_FREQ_96000,
-	},
-	.capabilities_size = sizeof(a2dp_ldac_t),
-	.channels[0] = a2dp_ldac_channels,
-	.channels_size[0] = ARRAYSIZE(a2dp_ldac_channels),
-	.samplings[0] = a2dp_ldac_samplings,
-	.samplings_size[0] = ARRAYSIZE(a2dp_ldac_samplings),
-};
-
-void a2dp_ldac_init(void) {
+static void a2dp_ldac_caps_intersect(
+		void *capabilities,
+		const void *mask) {
+	a2dp_caps_bitwise_intersect(capabilities, mask, sizeof(a2dp_ldac_t));
 }
 
-void a2dp_ldac_transport_init(struct ba_transport *t) {
-
-	const struct a2dp_codec *codec = t->a2dp.codec;
-
-	/* LDAC library internally for encoding uses 31-bit integers or
-	 * floats, so the best choice for PCM sample is signed 32-bit. */
-	t->a2dp.pcm.format = BA_TRANSPORT_PCM_FORMAT_S32_4LE;
-
-	t->a2dp.pcm.channels = a2dp_codec_lookup_channels(codec,
-			t->a2dp.configuration.ldac.channel_mode, false);
-	t->a2dp.pcm.sampling = a2dp_codec_lookup_frequency(codec,
-			t->a2dp.configuration.ldac.frequency, false);
-
+static int a2dp_ldac_caps_foreach_channel_mode(
+		const void *capabilities,
+		enum a2dp_stream stream,
+		a2dp_bit_mapping_foreach_func func,
+		void *userdata) {
+	const a2dp_ldac_t *caps = capabilities;
+	if (stream == A2DP_MAIN)
+		return a2dp_bit_mapping_foreach(a2dp_ldac_channels, caps->channel_mode, func, userdata);
+	return -1;
 }
 
-void *a2dp_ldac_enc_thread(struct ba_transport_thread *th) {
+static int a2dp_ldac_caps_foreach_sampling_freq(
+		const void *capabilities,
+		enum a2dp_stream stream,
+		a2dp_bit_mapping_foreach_func func,
+		void *userdata) {
+	const a2dp_ldac_t *caps = capabilities;
+	if (stream == A2DP_MAIN)
+		return a2dp_bit_mapping_foreach(a2dp_ldac_samplings, caps->sampling_freq, func, userdata);
+	return -1;
+}
+
+static void a2dp_ldac_caps_select_channel_mode(
+		void *capabilities,
+		enum a2dp_stream stream,
+		unsigned int channels) {
+	a2dp_ldac_t *caps = capabilities;
+	if (stream == A2DP_MAIN)
+		caps->channel_mode = a2dp_bit_mapping_lookup_value(a2dp_ldac_channels,
+				caps->channel_mode, channels);
+}
+
+static void a2dp_ldac_caps_select_sampling_freq(
+		void *capabilities,
+		enum a2dp_stream stream,
+		unsigned int frequency) {
+	a2dp_ldac_t *caps = capabilities;
+	if (stream == A2DP_MAIN)
+		caps->sampling_freq = a2dp_bit_mapping_lookup_value(a2dp_ldac_samplings,
+				caps->sampling_freq, frequency);
+}
+
+static struct a2dp_caps_helpers a2dp_ldac_caps_helpers = {
+	.intersect = a2dp_ldac_caps_intersect,
+	.has_stream = a2dp_caps_has_main_stream_only,
+	.foreach_channel_mode = a2dp_ldac_caps_foreach_channel_mode,
+	.foreach_sampling_freq = a2dp_ldac_caps_foreach_sampling_freq,
+	.select_channel_mode = a2dp_ldac_caps_select_channel_mode,
+	.select_sampling_freq = a2dp_ldac_caps_select_sampling_freq,
+};
+
+void *a2dp_ldac_enc_thread(struct ba_transport_pcm *t_pcm) {
 
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
-	pthread_cleanup_push(PTHREAD_CLEANUP(ba_transport_thread_cleanup), th);
+	pthread_cleanup_push(PTHREAD_CLEANUP(ba_transport_pcm_thread_cleanup), t_pcm);
 
-	struct ba_transport *t = th->t;
-	struct ba_transport_pcm *t_pcm = th->pcm;
+	struct ba_transport *t = t_pcm->t;
 	struct io_poll io = { .timeout = -1 };
 
 	HANDLE_LDAC_BT handle;
@@ -182,11 +176,10 @@ void *a2dp_ldac_enc_thread(struct ba_transport_thread *th) {
 	/* RTP clock frequency equal to audio samplerate */
 	rtp_state_init(&rtp, samplerate, samplerate);
 
-	debug_transport_thread_loop(th, "START");
-	for (ba_transport_thread_state_set_running(th);;) {
+	debug_transport_pcm_thread_loop(t_pcm, "START");
+	for (ba_transport_pcm_state_set_running(t_pcm);;) {
 
-		ssize_t samples = ffb_len_in(&pcm);
-		switch (samples = io_poll_and_read_pcm(&io, t_pcm, pcm.tail, samples)) {
+		switch (io_poll_and_read_pcm(&io, t_pcm, &pcm)) {
 		case -1:
 			if (errno == ESTALE) {
 				int tmp;
@@ -202,10 +195,8 @@ void *a2dp_ldac_enc_thread(struct ba_transport_thread *th) {
 			continue;
 		}
 
-		ffb_seek(&pcm, samples);
-		samples = ffb_len_out(&pcm);
-
 		int16_t *input = pcm.data;
+		size_t samples = ffb_len_out(&pcm);
 		size_t input_len = samples;
 
 		/* encode and transfer obtained data */
@@ -243,7 +234,7 @@ void *a2dp_ldac_enc_thread(struct ba_transport_thread *th) {
 				errno = 0;
 
 				ssize_t len = ffb_blen_out(&bt);
-				if ((len = io_bt_write(th, bt.data, len)) <= 0) {
+				if ((len = io_bt_write(t_pcm, bt.data, len)) <= 0) {
 					if (len == -1)
 						error("BT write error: %s", strerror(errno));
 					goto fail;
@@ -280,7 +271,7 @@ void *a2dp_ldac_enc_thread(struct ba_transport_thread *th) {
 	}
 
 fail:
-	debug_transport_thread_loop(th, "EXIT");
+	debug_transport_pcm_thread_loop(t_pcm, "EXIT");
 fail_ffb:
 	pthread_cleanup_pop(1);
 	pthread_cleanup_pop(1);
@@ -295,13 +286,12 @@ fail_open_ldac:
 
 #if HAVE_LDAC_DECODE
 __attribute__ ((weak))
-void *a2dp_ldac_dec_thread(struct ba_transport_thread *th) {
+void *a2dp_ldac_dec_thread(struct ba_transport_pcm *t_pcm) {
 
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
-	pthread_cleanup_push(PTHREAD_CLEANUP(ba_transport_thread_cleanup), th);
+	pthread_cleanup_push(PTHREAD_CLEANUP(ba_transport_pcm_thread_cleanup), t_pcm);
 
-	struct ba_transport *t = th->t;
-	struct ba_transport_pcm *t_pcm = th->pcm;
+	struct ba_transport *t = t_pcm->t;
 	struct io_poll io = { .timeout = -1 };
 
 	HANDLE_LDAC_BT handle;
@@ -337,11 +327,12 @@ void *a2dp_ldac_dec_thread(struct ba_transport_thread *th) {
 	/* RTP clock frequency equal to audio samplerate */
 	rtp_state_init(&rtp, samplerate, samplerate);
 
-	debug_transport_thread_loop(th, "START");
-	for (ba_transport_thread_state_set_running(th);;) {
+	debug_transport_pcm_thread_loop(t_pcm, "START");
+	for (ba_transport_pcm_state_set_running(t_pcm);;) {
 
-		ssize_t len = ffb_blen_in(&bt);
-		if ((len = io_poll_and_read_bt(&io, th, bt.data, len)) <= 0) {
+		ssize_t len;
+		ffb_rewind(&bt);
+		if ((len = io_poll_and_read_bt(&io, t_pcm, &bt)) <= 0) {
 			if (len == -1)
 				error("BT poll and read error: %s", strerror(errno));
 			goto fail;
@@ -381,7 +372,7 @@ void *a2dp_ldac_dec_thread(struct ba_transport_thread *th) {
 			const size_t samples = decoded / sample_size;
 			io_pcm_scale(t_pcm, pcm.data, samples);
 			if (io_pcm_write(t_pcm, pcm.data, samples) == -1)
-				error("FIFO write error: %s", strerror(errno));
+				error("PCM write error: %s", strerror(errno));
 
 			/* update local state with decoded PCM frames */
 			rtp_state_update(&rtp, samples / channels);
@@ -391,7 +382,7 @@ void *a2dp_ldac_dec_thread(struct ba_transport_thread *th) {
 	}
 
 fail:
-	debug_transport_thread_loop(th, "EXIT");
+	debug_transport_pcm_thread_loop(t_pcm, "EXIT");
 fail_ffb:
 	pthread_cleanup_pop(1);
 	pthread_cleanup_pop(1);
@@ -403,16 +394,154 @@ fail_open:
 }
 #endif
 
-int a2dp_ldac_transport_start(struct ba_transport *t) {
+static int a2dp_ldac_configuration_select(
+		const struct a2dp_sep *sep,
+		void *capabilities) {
 
-	if (t->profile & BA_TRANSPORT_PROFILE_A2DP_SOURCE)
-		return ba_transport_thread_create(&t->thread_enc, a2dp_ldac_enc_thread, "ba-a2dp-ldac", true);
+	a2dp_ldac_t *caps = capabilities;
+	const a2dp_ldac_t saved = *caps;
+
+	/* Narrow capabilities to values supported by BlueALSA. */
+	a2dp_ldac_caps_intersect(caps, &sep->config.capabilities);
+
+	unsigned int sampling_freq = 0;
+	if (a2dp_ldac_caps_foreach_sampling_freq(caps, A2DP_MAIN,
+				a2dp_bit_mapping_foreach_get_best_sampling_freq, &sampling_freq) != -1)
+		caps->sampling_freq = sampling_freq;
+	else {
+		error("LDAC: No supported sampling frequencies: %#x", saved.sampling_freq);
+		return errno = ENOTSUP, -1;
+	}
+
+	unsigned int channel_mode = 0;
+	if (a2dp_ldac_caps_foreach_channel_mode(caps, A2DP_MAIN,
+				a2dp_bit_mapping_foreach_get_best_channel_mode, &channel_mode) != -1)
+		caps->channel_mode = channel_mode;
+	else {
+		error("LDAC: No supported channel modes: %#x", saved.channel_mode);
+		return errno = ENOTSUP, -1;
+	}
+
+	return 0;
+}
+
+static int a2dp_ldac_configuration_check(
+		const struct a2dp_sep *sep,
+		const void *configuration) {
+
+	const a2dp_ldac_t *conf = configuration;
+	a2dp_ldac_t conf_v = *conf;
+
+	/* Validate configuration against BlueALSA capabilities. */
+	a2dp_ldac_caps_intersect(&conf_v, &sep->config.capabilities);
+
+	if (a2dp_bit_mapping_lookup(a2dp_ldac_samplings, conf_v.sampling_freq) == 0) {
+		debug("LDAC: Invalid sampling frequency: %#x", conf->sampling_freq);
+		return A2DP_CHECK_ERR_SAMPLING;
+	}
+
+	if (a2dp_bit_mapping_lookup(a2dp_ldac_channels, conf_v.channel_mode) == 0) {
+		debug("LDAC: Invalid channel mode: %#x", conf->channel_mode);
+		return A2DP_CHECK_ERR_CHANNEL_MODE;
+	}
+
+	return A2DP_CHECK_OK;
+}
+
+static int a2dp_ldac_transport_init(struct ba_transport *t) {
+
+	unsigned int channels;
+	if ((channels = a2dp_bit_mapping_lookup(a2dp_ldac_channels,
+					t->a2dp.configuration.ldac.channel_mode)) == 0)
+		return -1;
+
+	unsigned int sampling;
+	if ((sampling = a2dp_bit_mapping_lookup(a2dp_ldac_samplings,
+					t->a2dp.configuration.ldac.sampling_freq)) == 0)
+		return -1;
+
+	/* LDAC library internally for encoding uses 31-bit integers or
+	 * floats, so the best choice for PCM sample is signed 32-bit. */
+	t->a2dp.pcm.format = BA_TRANSPORT_PCM_FORMAT_S32_4LE;
+	t->a2dp.pcm.channels = channels;
+	t->a2dp.pcm.sampling = sampling;
+
+	return 0;
+}
+
+static int a2dp_ldac_source_init(struct a2dp_sep *sep) {
+	if (config.a2dp.force_mono)
+		sep->config.capabilities.ldac.channel_mode = LDAC_CHANNEL_MODE_MONO;
+	if (config.a2dp.force_44100)
+		sep->config.capabilities.ldac.sampling_freq = LDAC_SAMPLING_FREQ_44100;
+	return 0;
+}
+
+static int a2dp_ldac_source_transport_start(struct ba_transport *t) {
+	return ba_transport_pcm_start(&t->a2dp.pcm, a2dp_ldac_enc_thread, "ba-a2dp-ldac");
+}
+
+struct a2dp_sep a2dp_ldac_source = {
+	.name = "A2DP Source (LDAC)",
+	.config = {
+		.type = A2DP_SOURCE,
+		.codec_id = A2DP_CODEC_VENDOR_ID(LDAC_VENDOR_ID, LDAC_CODEC_ID),
+		.caps_size = sizeof(a2dp_ldac_t),
+		.capabilities.ldac = {
+			.info = A2DP_VENDOR_INFO_INIT(LDAC_VENDOR_ID, LDAC_CODEC_ID),
+			.channel_mode =
+				LDAC_CHANNEL_MODE_MONO |
+				LDAC_CHANNEL_MODE_DUAL |
+				LDAC_CHANNEL_MODE_STEREO,
+			/* NOTE: Used LDAC library does not support
+			 *       frequencies higher than 96 kHz. */
+			.sampling_freq =
+				LDAC_SAMPLING_FREQ_44100 |
+				LDAC_SAMPLING_FREQ_48000 |
+				LDAC_SAMPLING_FREQ_88200 |
+				LDAC_SAMPLING_FREQ_96000,
+		},
+	},
+	.init = a2dp_ldac_source_init,
+	.configuration_select = a2dp_ldac_configuration_select,
+	.configuration_check = a2dp_ldac_configuration_check,
+	.transport_init = a2dp_ldac_transport_init,
+	.transport_start = a2dp_ldac_source_transport_start,
+	.caps_helpers = &a2dp_ldac_caps_helpers,
+};
 
 #if HAVE_LDAC_DECODE
-	if (t->profile & BA_TRANSPORT_PROFILE_A2DP_SINK)
-		return ba_transport_thread_create(&t->thread_dec, a2dp_ldac_dec_thread, "ba-a2dp-ldac", true);
-#endif
 
-	g_assert_not_reached();
-	return -1;
+static int a2dp_ldac_sink_transport_start(struct ba_transport *t) {
+	return ba_transport_pcm_start(&t->a2dp.pcm, a2dp_ldac_dec_thread, "ba-a2dp-ldac");
 }
+
+struct a2dp_sep a2dp_ldac_sink = {
+	.name = "A2DP Sink (LDAC)",
+	.config = {
+		.type = A2DP_SINK,
+		.codec_id = A2DP_CODEC_VENDOR_ID(LDAC_VENDOR_ID, LDAC_CODEC_ID),
+		.caps_size = sizeof(a2dp_ldac_t),
+		.capabilities.ldac = {
+			.info = A2DP_VENDOR_INFO_INIT(LDAC_VENDOR_ID, LDAC_CODEC_ID),
+			.channel_mode =
+				LDAC_CHANNEL_MODE_MONO |
+				LDAC_CHANNEL_MODE_DUAL |
+				LDAC_CHANNEL_MODE_STEREO,
+			/* NOTE: Used LDAC library does not support
+			 *       frequencies higher than 96 kHz. */
+			.sampling_freq =
+				LDAC_SAMPLING_FREQ_44100 |
+				LDAC_SAMPLING_FREQ_48000 |
+				LDAC_SAMPLING_FREQ_88200 |
+				LDAC_SAMPLING_FREQ_96000,
+		},
+	},
+	.configuration_select = a2dp_ldac_configuration_select,
+	.configuration_check = a2dp_ldac_configuration_check,
+	.transport_init = a2dp_ldac_transport_init,
+	.transport_start = a2dp_ldac_sink_transport_start,
+	.caps_helpers = &a2dp_ldac_caps_helpers,
+};
+
+#endif
