@@ -1,6 +1,6 @@
 /*
  * BlueALSA - a2dp-mpeg.c
- * Copyright (c) 2016-2023 Arkadiusz Bokowy
+ * Copyright (c) 2016-2024 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -9,7 +9,10 @@
  */
 
 #include "a2dp-mpeg.h"
-/* IWYU pragma: no_include "config.h" */
+
+#if HAVE_CONFIG_H
+# include <config.h>
+#endif
 
 #include <errno.h>
 #include <pthread.h>
@@ -29,8 +32,9 @@
 #endif
 
 #include "a2dp.h"
+#include "ba-config.h"
 #include "ba-transport.h"
-#include "bluealsa-config.h"
+#include "ba-transport-pcm.h"
 #include "io.h"
 #include "rtp.h"
 #include "utils.h"
@@ -40,160 +44,88 @@
 #include "shared/log.h"
 #include "shared/rt.h"
 
-static const struct a2dp_channel_mode a2dp_mpeg_channels[] = {
-	{ A2DP_CHM_MONO, 1, MPEG_CHANNEL_MODE_MONO },
-	{ A2DP_CHM_DUAL_CHANNEL, 2, MPEG_CHANNEL_MODE_DUAL_CHANNEL },
-	{ A2DP_CHM_STEREO, 2, MPEG_CHANNEL_MODE_STEREO },
-	{ A2DP_CHM_JOINT_STEREO, 2, MPEG_CHANNEL_MODE_JOINT_STEREO },
+static const struct a2dp_bit_mapping a2dp_mpeg_channels[] = {
+	{ MPEG_CHANNEL_MODE_MONO, 1 },
+	{ MPEG_CHANNEL_MODE_DUAL_CHANNEL, 2 },
+	{ MPEG_CHANNEL_MODE_STEREO, 2 },
+	{ MPEG_CHANNEL_MODE_JOINT_STEREO, 2 },
+	{ 0 }
 };
 
-static const struct a2dp_sampling_freq a2dp_mpeg_samplings[] = {
-	{ 16000, MPEG_SAMPLING_FREQ_16000 },
-	{ 22050, MPEG_SAMPLING_FREQ_22050 },
-	{ 24000, MPEG_SAMPLING_FREQ_24000 },
-	{ 32000, MPEG_SAMPLING_FREQ_32000 },
-	{ 44100, MPEG_SAMPLING_FREQ_44100 },
-	{ 48000, MPEG_SAMPLING_FREQ_48000 },
+static const struct a2dp_bit_mapping a2dp_mpeg_samplings[] = {
+	{ MPEG_SAMPLING_FREQ_16000, 16000 },
+	{ MPEG_SAMPLING_FREQ_22050, 22050 },
+	{ MPEG_SAMPLING_FREQ_24000, 24000 },
+	{ MPEG_SAMPLING_FREQ_32000, 32000 },
+	{ MPEG_SAMPLING_FREQ_44100, 44100 },
+	{ MPEG_SAMPLING_FREQ_48000, 48000 },
+	{ 0 }
 };
 
-struct a2dp_codec a2dp_mpeg_sink = {
-	.dir = A2DP_SINK,
-	.codec_id = A2DP_CODEC_MPEG12,
-	.capabilities.mpeg = {
-		.layer =
-#if ENABLE_MPG123
-			MPEG_LAYER_MP1 |
-			MPEG_LAYER_MP2 |
-#endif
-			MPEG_LAYER_MP3,
-		.crc = 1,
-		/* NOTE: LAME does not support dual-channel mode. Be aware, that
-		 *       lack of this feature violates A2DP Sink specification. */
-		.channel_mode =
-			MPEG_CHANNEL_MODE_MONO |
-#if ENABLE_MPG123
-			MPEG_CHANNEL_MODE_DUAL_CHANNEL |
-#endif
-			MPEG_CHANNEL_MODE_STEREO |
-			MPEG_CHANNEL_MODE_JOINT_STEREO,
-		/* NOTE: Since MPF-2 is not required for neither Sink
-		 *       nor Source, we are not going to support it. */
-		.mpf = 0,
-		.frequency =
-			MPEG_SAMPLING_FREQ_16000 |
-			MPEG_SAMPLING_FREQ_22050 |
-			MPEG_SAMPLING_FREQ_24000 |
-			MPEG_SAMPLING_FREQ_32000 |
-			MPEG_SAMPLING_FREQ_44100 |
-			MPEG_SAMPLING_FREQ_48000,
-		.vbr = 1,
-		MPEG_INIT_BITRATE(
-			MPEG_BIT_RATE_INDEX_0 |
-			MPEG_BIT_RATE_INDEX_1 |
-			MPEG_BIT_RATE_INDEX_2 |
-			MPEG_BIT_RATE_INDEX_3 |
-			MPEG_BIT_RATE_INDEX_4 |
-			MPEG_BIT_RATE_INDEX_5 |
-			MPEG_BIT_RATE_INDEX_6 |
-			MPEG_BIT_RATE_INDEX_7 |
-			MPEG_BIT_RATE_INDEX_8 |
-			MPEG_BIT_RATE_INDEX_9 |
-			MPEG_BIT_RATE_INDEX_10 |
-			MPEG_BIT_RATE_INDEX_11 |
-			MPEG_BIT_RATE_INDEX_12 |
-			MPEG_BIT_RATE_INDEX_13 |
-			MPEG_BIT_RATE_INDEX_14
-		)
-	},
-	.capabilities_size = sizeof(a2dp_mpeg_t),
-	.channels[0] = a2dp_mpeg_channels,
-	.channels_size[0] = ARRAYSIZE(a2dp_mpeg_channels),
-	.samplings[0] = a2dp_mpeg_samplings,
-	.samplings_size[0] = ARRAYSIZE(a2dp_mpeg_samplings),
-	/* TODO: This is an optional but covered by the A2DP spec codec,
-	 *       so it could be enabled by default. However, it does not
-	 *       work reliably enough (for now)... */
-	.enabled = false,
-};
-
-struct a2dp_codec a2dp_mpeg_source = {
-	.dir = A2DP_SOURCE,
-	.codec_id = A2DP_CODEC_MPEG12,
-	.capabilities.mpeg = {
-		.layer =
-			MPEG_LAYER_MP3,
-		.crc = 1,
-		/* NOTE: LAME does not support dual-channel mode. */
-		.channel_mode =
-			MPEG_CHANNEL_MODE_MONO |
-			MPEG_CHANNEL_MODE_STEREO |
-			MPEG_CHANNEL_MODE_JOINT_STEREO,
-		/* NOTE: Since MPF-2 is not required for neither Sink
-		 *       nor Source, we are not going to support it. */
-		.mpf = 0,
-		.frequency =
-			MPEG_SAMPLING_FREQ_16000 |
-			MPEG_SAMPLING_FREQ_22050 |
-			MPEG_SAMPLING_FREQ_24000 |
-			MPEG_SAMPLING_FREQ_32000 |
-			MPEG_SAMPLING_FREQ_44100 |
-			MPEG_SAMPLING_FREQ_48000,
-		.vbr = 1,
-		MPEG_INIT_BITRATE(
-			MPEG_BIT_RATE_INDEX_0 |
-			MPEG_BIT_RATE_INDEX_1 |
-			MPEG_BIT_RATE_INDEX_2 |
-			MPEG_BIT_RATE_INDEX_3 |
-			MPEG_BIT_RATE_INDEX_4 |
-			MPEG_BIT_RATE_INDEX_5 |
-			MPEG_BIT_RATE_INDEX_6 |
-			MPEG_BIT_RATE_INDEX_7 |
-			MPEG_BIT_RATE_INDEX_8 |
-			MPEG_BIT_RATE_INDEX_9 |
-			MPEG_BIT_RATE_INDEX_10 |
-			MPEG_BIT_RATE_INDEX_11 |
-			MPEG_BIT_RATE_INDEX_12 |
-			MPEG_BIT_RATE_INDEX_13 |
-			MPEG_BIT_RATE_INDEX_14
-		)
-	},
-	.capabilities_size = sizeof(a2dp_mpeg_t),
-	.channels[0] = a2dp_mpeg_channels,
-	.channels_size[0] = ARRAYSIZE(a2dp_mpeg_channels),
-	.samplings[0] = a2dp_mpeg_samplings,
-	.samplings_size[0] = ARRAYSIZE(a2dp_mpeg_samplings),
-	.enabled = false,
-};
-
-void a2dp_mpeg_init(void) {
-
-	if (config.a2dp.force_mono)
-		a2dp_mpeg_source.capabilities.mpeg.channel_mode = MPEG_CHANNEL_MODE_MONO;
-	if (config.a2dp.force_44100)
-		a2dp_mpeg_source.capabilities.mpeg.frequency = MPEG_SAMPLING_FREQ_44100;
-
+static void a2dp_mpeg_caps_intersect(
+		void *capabilities,
+		const void *mask) {
+	a2dp_caps_bitwise_intersect(capabilities, mask, sizeof(a2dp_mpeg_t));
 }
 
-void a2dp_mpeg_transport_init(struct ba_transport *t) {
-
-	const struct a2dp_codec *codec = t->a2dp.codec;
-
-	t->a2dp.pcm.format = BA_TRANSPORT_PCM_FORMAT_S16_2LE;
-	t->a2dp.pcm.channels = a2dp_codec_lookup_channels(codec,
-			t->a2dp.configuration.mpeg.channel_mode, false);
-	t->a2dp.pcm.sampling = a2dp_codec_lookup_frequency(codec,
-			t->a2dp.configuration.mpeg.frequency, false);
-
+static int a2dp_mpeg_caps_foreach_channel_mode(
+		const void *capabilities,
+		enum a2dp_stream stream,
+		a2dp_bit_mapping_foreach_func func,
+		void *userdata) {
+	const a2dp_mpeg_t *caps = capabilities;
+	if (stream == A2DP_MAIN)
+		return a2dp_bit_mapping_foreach(a2dp_mpeg_channels, caps->channel_mode, func, userdata);
+	return -1;
 }
 
+static int a2dp_mpeg_caps_foreach_sampling_freq(
+		const void *capabilities,
+		enum a2dp_stream stream,
+		a2dp_bit_mapping_foreach_func func,
+		void *userdata) {
+	const a2dp_mpeg_t *caps = capabilities;
+	if (stream == A2DP_MAIN)
+		return a2dp_bit_mapping_foreach(a2dp_mpeg_samplings, caps->sampling_freq, func, userdata);
+	return -1;
+}
+
+static void a2dp_mpeg_caps_select_channel_mode(
+		void *capabilities,
+		enum a2dp_stream stream,
+		unsigned int channels) {
+	a2dp_mpeg_t *caps = capabilities;
+	if (stream == A2DP_MAIN)
+		caps->channel_mode = a2dp_bit_mapping_lookup_value(a2dp_mpeg_channels,
+				caps->channel_mode, channels);
+}
+
+static void a2dp_mpeg_caps_select_sampling_freq(
+		void *capabilities,
+		enum a2dp_stream stream,
+		unsigned int frequency) {
+	a2dp_mpeg_t *caps = capabilities;
+	if (stream == A2DP_MAIN)
+		caps->sampling_freq = a2dp_bit_mapping_lookup_value(a2dp_mpeg_samplings,
+				caps->sampling_freq, frequency);
+}
+
+static struct a2dp_caps_helpers a2dp_mpeg_caps_helpers = {
+	.intersect = a2dp_mpeg_caps_intersect,
+	.has_stream = a2dp_caps_has_main_stream_only,
+	.foreach_channel_mode = a2dp_mpeg_caps_foreach_channel_mode,
+	.foreach_sampling_freq = a2dp_mpeg_caps_foreach_sampling_freq,
+	.select_channel_mode = a2dp_mpeg_caps_select_channel_mode,
+	.select_sampling_freq = a2dp_mpeg_caps_select_sampling_freq,
+};
+
 #if ENABLE_MP3LAME
-void *a2dp_mp3_enc_thread(struct ba_transport_thread *th) {
+void *a2dp_mp3_enc_thread(struct ba_transport_pcm *t_pcm) {
 
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
-	pthread_cleanup_push(PTHREAD_CLEANUP(ba_transport_thread_cleanup), th);
+	pthread_cleanup_push(PTHREAD_CLEANUP(ba_transport_pcm_thread_cleanup), t_pcm);
 
-	struct ba_transport *t = th->t;
-	struct ba_transport_pcm *t_pcm = th->pcm;
+	struct ba_transport *t = t_pcm->t;
 	struct io_poll io = { .timeout = -1 };
 
 	lame_t handle;
@@ -254,13 +186,13 @@ void *a2dp_mp3_enc_thread(struct ba_transport_thread *th) {
 			error("LAME: Couldn't set CBR mode");
 			goto fail_setup;
 		}
-		int mpeg_bitrate = MPEG_GET_BITRATE(*configuration);
+		int mpeg_bitrate = A2DP_MPEG_GET_BITRATE(*configuration);
 		int bitrate = a2dp_mpeg1_mp3_get_max_bitrate(mpeg_bitrate);
 		if (lame_set_brate(handle, bitrate) != 0) {
 			error("LAME: Couldn't set CBR bitrate: %d", bitrate);
 			goto fail_setup;
 		}
-		if (mpeg_bitrate & MPEG_BIT_RATE_FREE &&
+		if (mpeg_bitrate & MPEG_BITRATE_FREE &&
 				lame_set_free_format(handle, 1) != 0) {
 			error("LAME: Couldn't enable free format");
 			goto fail_setup;
@@ -304,11 +236,10 @@ void *a2dp_mp3_enc_thread(struct ba_transport_thread *th) {
 	/* RTP clock frequency equal to 90kHz */
 	rtp_state_init(&rtp, samplerate, 90000);
 
-	debug_transport_thread_loop(th, "START");
-	for (ba_transport_thread_state_set_running(th);;) {
+	debug_transport_pcm_thread_loop(t_pcm, "START");
+	for (ba_transport_pcm_state_set_running(t_pcm);;) {
 
-		ssize_t samples = ffb_len_in(&pcm);
-		switch (samples = io_poll_and_read_pcm(&io, t_pcm, pcm.tail, samples)) {
+		switch (io_poll_and_read_pcm(&io, t_pcm, &pcm)) {
 		case -1:
 			if (errno == ESTALE) {
 				lame_encode_flush(handle, rtp_payload, mpeg_frame_len);
@@ -322,13 +253,10 @@ void *a2dp_mp3_enc_thread(struct ba_transport_thread *th) {
 			continue;
 		}
 
-		ffb_seek(&pcm, samples);
-		samples = ffb_len_out(&pcm);
-
 		/* anchor for RTP payload */
 		bt.tail = rtp_payload;
 
-		size_t pcm_frames = samples / channels;
+		size_t pcm_frames = ffb_len_out(&pcm) / channels;
 		ssize_t len;
 
 		if ((len = channels == 1 ?
@@ -356,7 +284,7 @@ void *a2dp_mp3_enc_thread(struct ba_transport_thread *th) {
 				ffb_seek(&bt, RTP_HEADER_LEN + sizeof(*rtp_mpeg_audio_header) + chunk_len);
 
 				len = ffb_blen_out(&bt);
-				if ((len = io_bt_write(th, bt.data, len)) <= 0) {
+				if ((len = io_bt_write(t_pcm, bt.data, len)) <= 0) {
 					if (len == -1)
 						error("BT write error: %s", strerror(errno));
 					goto fail;
@@ -394,7 +322,7 @@ void *a2dp_mp3_enc_thread(struct ba_transport_thread *th) {
 	}
 
 fail:
-	debug_transport_thread_loop(th, "EXIT");
+	debug_transport_pcm_thread_loop(t_pcm, "EXIT");
 fail_ffb:
 	pthread_cleanup_pop(1);
 	pthread_cleanup_pop(1);
@@ -406,15 +334,14 @@ fail_init:
 }
 #endif
 
-#if ENABLE_MP3LAME || ENABLE_MPG123
+#if ENABLE_MPG123 || ENABLE_MP3LAME
 __attribute__ ((weak))
-void *a2dp_mpeg_dec_thread(struct ba_transport_thread *th) {
+void *a2dp_mpeg_dec_thread(struct ba_transport_pcm *t_pcm) {
 
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
-	pthread_cleanup_push(PTHREAD_CLEANUP(ba_transport_thread_cleanup), th);
+	pthread_cleanup_push(PTHREAD_CLEANUP(ba_transport_pcm_thread_cleanup), t_pcm);
 
-	struct ba_transport *t = th->t;
-	struct ba_transport_pcm *t_pcm = th->pcm;
+	struct ba_transport *t = t_pcm->t;
 	struct io_poll io = { .timeout = -1 };
 
 #if ENABLE_MPG123
@@ -488,11 +415,12 @@ void *a2dp_mpeg_dec_thread(struct ba_transport_thread *th) {
 	/* RTP clock frequency equal to 90kHz */
 	rtp_state_init(&rtp, samplerate, 90000);
 
-	debug_transport_thread_loop(th, "START");
-	for (ba_transport_thread_state_set_running(th);;) {
+	debug_transport_pcm_thread_loop(t_pcm, "START");
+	for (ba_transport_pcm_state_set_running(t_pcm);;) {
 
-		ssize_t len = ffb_blen_in(&bt);
-		if ((len = io_poll_and_read_bt(&io, th, bt.data, len)) <= 0) {
+		ssize_t len;
+		ffb_rewind(&bt);
+		if ((len = io_poll_and_read_bt(&io, t_pcm, &bt)) <= 0) {
 			if (len == -1)
 				error("BT poll and read error: %s", strerror(errno));
 			goto fail;
@@ -539,7 +467,7 @@ decode:
 		const size_t samples = len / sizeof(int16_t);
 		io_pcm_scale(t_pcm, pcm.data, samples);
 		if (io_pcm_write(t_pcm, pcm.data, samples) == -1)
-			error("FIFO write error: %s", strerror(errno));
+			error("PCM write error: %s", strerror(errno));
 
 		/* update local state with decoded PCM frames */
 		rtp_state_update(&rtp, samples / channels);
@@ -563,19 +491,18 @@ decode:
 		if (channels == 1) {
 			io_pcm_scale(t_pcm, pcm_l, samples);
 			if (io_pcm_write(t_pcm, pcm_l, samples) == -1)
-				error("FIFO write error: %s", strerror(errno));
+				error("PCM write error: %s", strerror(errno));
 		}
 		else {
 
-			ssize_t i;
-			for (i = 0; i < samples; i++) {
+			for (ssize_t i = 0; i < samples; i++) {
 				((int16_t *)pcm.data)[i * 2 + 0] = pcm_l[i];
 				((int16_t *)pcm.data)[i * 2 + 1] = pcm_r[i];
 			}
 
 			io_pcm_scale(t_pcm, pcm.data, samples);
 			if (io_pcm_write(t_pcm, pcm.data, samples) == -1)
-				error("FIFO write error: %s", strerror(errno));
+				error("PCM write error: %s", strerror(errno));
 
 		}
 
@@ -587,7 +514,7 @@ decode:
 	}
 
 fail:
-	debug_transport_thread_loop(th, "EXIT");
+	debug_transport_pcm_thread_loop(t_pcm, "EXIT");
 fail_ffb:
 	pthread_cleanup_pop(1);
 	pthread_cleanup_pop(1);
@@ -601,24 +528,253 @@ fail_init:
 }
 #endif
 
-int a2dp_mpeg_transport_start(struct ba_transport *t) {
+static int a2dp_mpeg_configuration_select(
+		const struct a2dp_sep *sep,
+		void *capabilities) {
+
+	a2dp_mpeg_t *caps = capabilities;
+	const a2dp_mpeg_t saved = *caps;
+
+	/* Narrow capabilities to values supported by BlueALSA. */
+	a2dp_mpeg_caps_intersect(caps, &sep->config.capabilities);
+
+	if (caps->layer & MPEG_LAYER_MP3)
+		caps->layer = MPEG_LAYER_MP3;
+	else if (caps->layer & MPEG_LAYER_MP2)
+		caps->layer = MPEG_LAYER_MP2;
+	else if (caps->layer & MPEG_LAYER_MP1)
+		caps->layer = MPEG_LAYER_MP1;
+	else {
+		error("MPEG: No supported layers: %#x", saved.layer);
+		return errno = ENOTSUP, -1;
+	}
+
+	unsigned int channel_mode = 0;
+	if (a2dp_mpeg_caps_foreach_channel_mode(caps, A2DP_MAIN,
+				a2dp_bit_mapping_foreach_get_best_channel_mode, &channel_mode) != -1)
+		caps->channel_mode = channel_mode;
+	else {
+		error("MPEG: No supported channel modes: %#x", saved.channel_mode);
+		return errno = ENOTSUP, -1;
+	}
+
+	unsigned int sampling_freq = 0;
+	if (a2dp_mpeg_caps_foreach_sampling_freq(caps, A2DP_MAIN,
+				a2dp_bit_mapping_foreach_get_best_sampling_freq, &sampling_freq) != -1)
+		caps->sampling_freq = sampling_freq;
+	else {
+		error("MPEG: No supported sampling frequencies: %#x", saved.sampling_freq);
+		return errno = ENOTSUP, -1;
+	}
+
+	/* do not waste bits for CRC protection */
+	caps->crc = 0;
+	/* do not use MPF-2 */
+	caps->mpf = 0;
+
+	return 0;
+}
+
+static int a2dp_mpeg_configuration_check(
+		const struct a2dp_sep *sep,
+		const void *configuration) {
+
+	const a2dp_mpeg_t *conf = configuration;
+	a2dp_mpeg_t conf_v = *conf;
+
+	/* Validate configuration against BlueALSA capabilities. */
+	a2dp_mpeg_caps_intersect(&conf_v, &sep->config.capabilities);
+
+	switch (conf_v.layer) {
+	case MPEG_LAYER_MP1:
+	case MPEG_LAYER_MP2:
+	case MPEG_LAYER_MP3:
+		break;
+	default:
+		debug("MPEG: Invalid layer: %#x", conf->layer);
+		return A2DP_CHECK_ERR_MPEG_LAYER;
+	}
+
+	if (a2dp_bit_mapping_lookup(a2dp_mpeg_channels, conf_v.channel_mode) == 0) {
+		debug("MPEG: Invalid channel mode: %#x", conf->channel_mode);
+		return A2DP_CHECK_ERR_CHANNEL_MODE;
+	}
+
+	if (a2dp_bit_mapping_lookup(a2dp_mpeg_samplings, conf_v.sampling_freq) == 0) {
+		debug("MPEG: Invalid sampling frequency: %#x", conf->sampling_freq);
+		return A2DP_CHECK_ERR_SAMPLING;
+	}
+
+	return A2DP_CHECK_OK;
+}
+
+static int a2dp_mpeg_transport_init(struct ba_transport *t) {
+
+	unsigned int channels;
+	if ((channels = a2dp_bit_mapping_lookup(a2dp_mpeg_channels,
+					t->a2dp.configuration.mpeg.channel_mode)) == 0)
+		return -1;
+
+	unsigned int sampling;
+	if ((sampling = a2dp_bit_mapping_lookup(a2dp_mpeg_samplings,
+					t->a2dp.configuration.mpeg.sampling_freq)) == 0)
+		return -1;
+
+	t->a2dp.pcm.format = BA_TRANSPORT_PCM_FORMAT_S16_2LE;
+	t->a2dp.pcm.channels = channels;
+	t->a2dp.pcm.sampling = sampling;
+
+	return 0;
+}
 
-	if (t->profile & BA_TRANSPORT_PROFILE_A2DP_SOURCE) {
 #if ENABLE_MP3LAME
-		if (t->a2dp.configuration.mpeg.layer == MPEG_LAYER_MP3)
-			return ba_transport_thread_create(&t->thread_enc, a2dp_mp3_enc_thread, "ba-a2dp-mp3", true);
-#endif
-	}
 
-	if (t->profile & BA_TRANSPORT_PROFILE_A2DP_SINK) {
-#if ENABLE_MPG123
-		return ba_transport_thread_create(&t->thread_dec, a2dp_mpeg_dec_thread, "ba-a2dp-mpeg", true);
-#elif ENABLE_MP3LAME
-		if (t->a2dp.configuration.mpeg.layer == MPEG_LAYER_MP3)
-			return ba_transport_thread_create(&t->thread_dec, a2dp_mpeg_dec_thread, "ba-a2dp-mp3", true);
-#endif
-	}
+static int a2dp_mpeg_source_init(struct a2dp_sep *sep) {
+	if (config.a2dp.force_mono)
+		sep->config.capabilities.mpeg.channel_mode = MPEG_CHANNEL_MODE_MONO;
+	if (config.a2dp.force_44100)
+		sep->config.capabilities.mpeg.sampling_freq = MPEG_SAMPLING_FREQ_44100;
+	return 0;
+}
 
+static int a2dp_mpeg_source_transport_start(struct ba_transport *t) {
+	if (t->a2dp.configuration.mpeg.layer == MPEG_LAYER_MP3)
+		return ba_transport_pcm_start(&t->a2dp.pcm, a2dp_mp3_enc_thread, "ba-a2dp-mp3");
 	g_assert_not_reached();
 	return -1;
 }
+
+struct a2dp_sep a2dp_mpeg_source = {
+	.name = "A2DP Source (MP3)",
+	.config = {
+		.type = A2DP_SOURCE,
+		.codec_id = A2DP_CODEC_MPEG12,
+		.caps_size = sizeof(a2dp_mpeg_t),
+		.capabilities.mpeg = {
+			.layer =
+				MPEG_LAYER_MP3,
+			.crc = 1,
+			/* NOTE: LAME does not support dual-channel mode. */
+			.channel_mode =
+				MPEG_CHANNEL_MODE_MONO |
+				MPEG_CHANNEL_MODE_STEREO |
+				MPEG_CHANNEL_MODE_JOINT_STEREO,
+			/* NOTE: Since MPF-2 is not required for neither Sink
+			 *       nor Source, we are not going to support it. */
+			.mpf = 0,
+			.sampling_freq =
+				MPEG_SAMPLING_FREQ_16000 |
+				MPEG_SAMPLING_FREQ_22050 |
+				MPEG_SAMPLING_FREQ_24000 |
+				MPEG_SAMPLING_FREQ_32000 |
+				MPEG_SAMPLING_FREQ_44100 |
+				MPEG_SAMPLING_FREQ_48000,
+			.vbr = 1,
+			A2DP_MPEG_INIT_BITRATE(
+				MPEG_BITRATE_INDEX_0 |
+				MPEG_BITRATE_INDEX_1 |
+				MPEG_BITRATE_INDEX_2 |
+				MPEG_BITRATE_INDEX_3 |
+				MPEG_BITRATE_INDEX_4 |
+				MPEG_BITRATE_INDEX_5 |
+				MPEG_BITRATE_INDEX_6 |
+				MPEG_BITRATE_INDEX_7 |
+				MPEG_BITRATE_INDEX_8 |
+				MPEG_BITRATE_INDEX_9 |
+				MPEG_BITRATE_INDEX_10 |
+				MPEG_BITRATE_INDEX_11 |
+				MPEG_BITRATE_INDEX_12 |
+				MPEG_BITRATE_INDEX_13 |
+				MPEG_BITRATE_INDEX_14
+			)
+		},
+	},
+	.init = a2dp_mpeg_source_init,
+	.configuration_select = a2dp_mpeg_configuration_select,
+	.configuration_check = a2dp_mpeg_configuration_check,
+	.transport_init = a2dp_mpeg_transport_init,
+	.transport_start = a2dp_mpeg_source_transport_start,
+	.caps_helpers = &a2dp_mpeg_caps_helpers,
+	/* TODO: This is an optional but covered by the A2DP spec codec,
+	 *       so it could be enabled by default. However, it does not
+	 *       work reliably enough (for now)... */
+	.enabled = false,
+};
+
+#endif
+
+#if ENABLE_MPG123 || ENABLE_MP3LAME
+
+static int a2dp_mpeg_sink_transport_start(struct ba_transport *t) {
+#if ENABLE_MPG123
+	return ba_transport_pcm_start(&t->a2dp.pcm, a2dp_mpeg_dec_thread, "ba-a2dp-mpeg");
+#else
+	if (t->a2dp.configuration.mpeg.layer == MPEG_LAYER_MP3)
+		return ba_transport_pcm_start(&t->a2dp.pcm, a2dp_mpeg_dec_thread, "ba-a2dp-mp3");
+	g_assert_not_reached();
+	return -1;
+#endif
+}
+
+struct a2dp_sep a2dp_mpeg_sink = {
+	.name = "A2DP Sink (MP3)",
+	.config = {
+		.type = A2DP_SINK,
+		.codec_id = A2DP_CODEC_MPEG12,
+		.caps_size = sizeof(a2dp_mpeg_t),
+		.capabilities.mpeg = {
+			.layer =
+#if ENABLE_MPG123
+				MPEG_LAYER_MP1 |
+				MPEG_LAYER_MP2 |
+#endif
+				MPEG_LAYER_MP3,
+			.crc = 1,
+			/* NOTE: LAME does not support dual-channel mode. Be aware, that
+			 *       lack of this feature violates A2DP Sink specification. */
+			.channel_mode =
+				MPEG_CHANNEL_MODE_MONO |
+#if ENABLE_MPG123
+				MPEG_CHANNEL_MODE_DUAL_CHANNEL |
+#endif
+				MPEG_CHANNEL_MODE_STEREO |
+				MPEG_CHANNEL_MODE_JOINT_STEREO,
+			/* NOTE: Since MPF-2 is not required for neither Sink
+			 *       nor Source, we are not going to support it. */
+			.mpf = 0,
+			.sampling_freq =
+				MPEG_SAMPLING_FREQ_16000 |
+				MPEG_SAMPLING_FREQ_22050 |
+				MPEG_SAMPLING_FREQ_24000 |
+				MPEG_SAMPLING_FREQ_32000 |
+				MPEG_SAMPLING_FREQ_44100 |
+				MPEG_SAMPLING_FREQ_48000,
+			.vbr = 1,
+			A2DP_MPEG_INIT_BITRATE(
+				MPEG_BITRATE_INDEX_0 |
+				MPEG_BITRATE_INDEX_1 |
+				MPEG_BITRATE_INDEX_2 |
+				MPEG_BITRATE_INDEX_3 |
+				MPEG_BITRATE_INDEX_4 |
+				MPEG_BITRATE_INDEX_5 |
+				MPEG_BITRATE_INDEX_6 |
+				MPEG_BITRATE_INDEX_7 |
+				MPEG_BITRATE_INDEX_8 |
+				MPEG_BITRATE_INDEX_9 |
+				MPEG_BITRATE_INDEX_10 |
+				MPEG_BITRATE_INDEX_11 |
+				MPEG_BITRATE_INDEX_12 |
+				MPEG_BITRATE_INDEX_13 |
+				MPEG_BITRATE_INDEX_14
+			)
+		},
+	},
+	.configuration_select = a2dp_mpeg_configuration_select,
+	.configuration_check = a2dp_mpeg_configuration_check,
+	.transport_init = a2dp_mpeg_transport_init,
+	.transport_start = a2dp_mpeg_sink_transport_start,
+	.caps_helpers = &a2dp_mpeg_caps_helpers,
+	.enabled = false,
+};
+
+#endif
