@@ -1,6 +1,6 @@
 /*
  * bluealsa-pcm.c
- * Copyright (c) 2016-2022 Arkadiusz Bokowy
+ * Copyright (c) 2016-2024 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -12,7 +12,6 @@
 # include <config.h>
 #endif
 
-#include <alloca.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <poll.h>
@@ -36,6 +35,7 @@
 #include <dbus/dbus.h>
 
 #include "shared/dbus-client.h"
+#include "shared/dbus-client-pcm.h"
 #include "shared/defs.h"
 #include "shared/hex.h"
 #include "shared/log.h"
@@ -45,11 +45,12 @@
 #define BA_PAUSE_STATE_PAUSED  (1 << 0)
 #define BA_PAUSE_STATE_PENDING (1 << 1)
 
-#if SND_LIB_VERSION >= 0x010104
+#if SND_LIB_VERSION >= 0x010104 && SND_LIB_VERSION < 0x010206
+#include <alloca.h>
 /**
- * alsa-lib releases from 1.1.4 have a bug in the rate plugin
- * which, when combined with the hw params refinement algorithm used by the
- * ioplug, can cause snd_pcm_avail() to return bogus values. This, in turn,
+ * alsa-lib releases from 1.1.4 to 1.2.5.1 inclusive have a bug in the rate
+ * plugin which, when combined with the hw params refinement algorithm used by
+ * the ioplug, can cause snd_pcm_avail() to return bogus values. This, in turn,
  * can trigger deadlock in applications built on the portaudio library
  * (e.g. audacity) and possibly cause faults in other applications too.
  *
@@ -71,6 +72,11 @@ struct bluealsa_pcm {
 
 	/* requested BlueALSA PCM */
 	struct ba_pcm ba_pcm;
+	/* user-provided codec configuration */
+	uint8_t ba_pcm_codec_config[64];
+	size_t ba_pcm_codec_config_len;
+	/* additional supported codecs */
+	struct ba_pcm_codecs ba_pcm_codecs;
 
 	/* PCM FIFO */
 	int ba_pcm_fd;
@@ -279,76 +285,88 @@ static void *io_thread(snd_pcm_ioplug_t *io) {
 		if (frames > avail)
 			frames = avail;
 
-		/* When used with the rate plugin the buffer might contain a fractional
-		 * number of periods. So if the leftover in the buffer is less than a
-		 * whole period size, adjust the number of frames which should be
-		 * transferred.  */
-		if (io->buffer_size - offset < frames)
-			frames = io->buffer_size - offset;
-
-		/* IO operation size in bytes */
-		size_t len = frames * pcm->frame_size;
-		char *head = pcm->io_hw_buffer + offset * pcm->frame_size;
-
 		/* Increment the HW pointer (with boundary wrap). */
 		io_hw_ptr += frames;
 		if ((snd_pcm_uframes_t)io_hw_ptr >= pcm->io_hw_boundary)
 			io_hw_ptr -= pcm->io_hw_boundary;
 
-		ssize_t ret = 0;
-		if (io->stream == SND_PCM_STREAM_CAPTURE) {
+		/* When used with the rate plugin the buffer size may not be an
+		 * integer multiple of the period size. If so, the current period may
+		 * be split, part at the end of the buffer and the remainder at the
+		 * start. In this case we must perform the transfer in two chunks to
+		 * make up a full period.  */
+		snd_pcm_uframes_t chunk = frames;
+		if (io->buffer_size - offset < frames)
+			chunk = io->buffer_size - offset;
 
-			/* Read the whole period "atomically". This will assure, that frames
-			 * are not fragmented, so the pointer can be correctly updated. */
-			while (len != 0 && (ret = read(pcm->ba_pcm_fd, head, len)) != 0) {
-				if (ret == -1) {
-					if (errno == EINTR)
-						continue;
-					SNDERR("PCM FIFO read error: %s", strerror(errno));
+		snd_pcm_uframes_t frames_transfered = 0;
+		while (frames_transfered < frames) {
+
+			/* IO operation size in bytes */
+			size_t len = chunk * pcm->frame_size;
+			char *head = pcm->io_hw_buffer + offset * pcm->frame_size;
+
+			ssize_t ret = 0;
+			if (io->stream == SND_PCM_STREAM_CAPTURE) {
+
+				/* Read the whole chunk "atomically". This will assure, that
+				 * frames are not fragmented, so the pointer can be correctly
+				 * updated. */
+				while (len != 0 && (ret = read(pcm->ba_pcm_fd, head, len)) != 0) {
+					if (ret == -1) {
+						if (errno == EINTR)
+							continue;
+						SNDERR("PCM FIFO read error: %s", strerror(errno));
+						pcm->connected = false;
+						goto fail;
+					}
+					head += ret;
+					len -= ret;
+				}
+
+				if (ret == 0) {
 					pcm->connected = false;
 					goto fail;
 				}
-				head += ret;
-				len -= ret;
+
+			}
+			else {
+
+				/* Perform atomic write - see the explanation above. */
+				do {
+					if ((ret = write(pcm->ba_pcm_fd, head, len)) == -1) {
+						if (errno == EINTR)
+							continue;
+						if (errno != EPIPE)
+							SNDERR("PCM FIFO write error: %s", strerror(errno));
+						pcm->connected = false;
+						goto fail;
+					}
+					head += ret;
+					len -= ret;
+				} while (len != 0);
+
 			}
 
-			if (ret == 0) {
-				pcm->connected = false;
-				goto fail;
-			}
-
-			io_thread_update_delay(pcm, io_hw_ptr);
+			frames_transfered += chunk;
+			offset = 0;
+			chunk = frames - chunk;
 
 		}
-		else {
 
-			/* Perform atomic write - see the explanation above. */
-			do {
-				if ((ret = write(pcm->ba_pcm_fd, head, len)) == -1) {
-					if (errno == EINTR)
-						continue;
-					if (errno != EPIPE)
-						SNDERR("PCM FIFO write error: %s", strerror(errno));
-					pcm->connected = false;
-					goto fail;
-				}
-				head += ret;
-				len -= ret;
-			} while (len != 0);
+		io_thread_update_delay(pcm, io_hw_ptr);
 
-			io_thread_update_delay(pcm, io_hw_ptr);
-
-			/* synchronize playback time */
+		/* synchronize playback time */
+		if (io->stream == SND_PCM_STREAM_PLAYBACK)
 			asrsync_sync(&asrs, frames);
 
-		}
-
 		/* Make the new HW pointer value visible to the ioplug. */
 		pcm->io_hw_ptr = io_hw_ptr;
 
 		/* Wake application thread if enough space/frames is available. */
 		if (frames + io->buffer_size - avail >= pcm->io_avail_min)
 			eventfd_write(pcm->event_fd, 1);
+
 	}
 
 fail:
@@ -381,9 +399,9 @@ static int bluealsa_start(snd_pcm_ioplug_t *io) {
 		return 0;
 	}
 
-	if (!bluealsa_dbus_pcm_ctrl_send_resume(pcm->ba_pcm_ctrl_fd, NULL)) {
+	if (!ba_dbus_pcm_ctrl_send_resume(pcm->ba_pcm_ctrl_fd, NULL)) {
 		debug2("Couldn't start PCM: %s", strerror(errno));
-		return -errno;
+		return -EIO;
 	}
 
 	/* Initialize delay calculation - capture reception begins immediately,
@@ -398,7 +416,7 @@ static int bluealsa_start(snd_pcm_ioplug_t *io) {
 					PTHREAD_FUNC(io_thread), io)) != 0) {
 		debug2("Couldn't create IO thread: %s", strerror(errno));
 		pcm->io_started = false;
-		return -errno;
+		return -EIO;
 	}
 
 	pthread_setname_np(pcm->io_thread, "pcm-io");
@@ -419,8 +437,8 @@ static int bluealsa_stop(snd_pcm_ioplug_t *io) {
 	 * was stopped. */
 	pcm->io_hw_ptr = 0;
 
-	if (!bluealsa_dbus_pcm_ctrl_send_drop(pcm->ba_pcm_ctrl_fd, NULL))
-		return -errno;
+	if (!ba_dbus_pcm_ctrl_send_drop(pcm->ba_pcm_ctrl_fd, NULL))
+		return -EIO;
 
 	/* Applications that call poll() after snd_pcm_drain() will be blocked
 	 * forever unless we generate a poll() event here. */
@@ -455,8 +473,10 @@ static snd_pcm_sframes_t bluealsa_pointer(snd_pcm_ioplug_t *io) {
 static int bluealsa_close(snd_pcm_ioplug_t *io) {
 	struct bluealsa_pcm *pcm = io->private_data;
 	debug2("Closing");
-	bluealsa_dbus_connection_ctx_free(&pcm->dbus_ctx);
-	close(pcm->event_fd);
+	ba_dbus_pcm_codecs_free(&pcm->ba_pcm_codecs);
+	ba_dbus_connection_ctx_free(&pcm->dbus_ctx);
+	if (pcm->event_fd != -1)
+		close(pcm->event_fd);
 	pthread_mutex_destroy(&pcm->mutex);
 	pthread_cond_destroy(&pcm->pause_cond);
 	free(pcm);
@@ -545,14 +565,41 @@ static int bluealsa_hw_params(snd_pcm_ioplug_t *io, snd_pcm_hw_params_t *params)
 
 	debug2("Initializing HW");
 
-#if	BLUEALSA_HW_PARAMS_FIX
+	DBusError err = DBUS_ERROR_INIT;
+	int ret;
+
+	unsigned int channels;
+	if ((ret = snd_pcm_hw_params_get_channels(params, &channels)) < 0)
+		return ret;
+
+	unsigned int sampling;
+	if ((ret = snd_pcm_hw_params_get_rate(params, &sampling, NULL)) < 0)
+		return ret;
+
+	if (pcm->ba_pcm.channels != channels || pcm->ba_pcm.sampling != sampling) {
+		debug2("Changing BlueALSA PCM configuration: %u ch, %u Hz -> %u ch, %u Hz",
+				pcm->ba_pcm.channels, pcm->ba_pcm.sampling, channels, sampling);
+
+		if (ba_dbus_pcm_select_codec(&pcm->dbus_ctx, pcm->ba_pcm.pcm_path,
+				pcm->ba_pcm.codec.name, pcm->ba_pcm_codec_config, pcm->ba_pcm_codec_config_len,
+				channels, sampling, BA_PCM_SELECT_CODEC_FLAG_NONE, &err)) {
+			pcm->ba_pcm.channels = channels;
+			pcm->ba_pcm.sampling = sampling;
+		}
+		else {
+			SNDERR("Couldn't change BlueALSA PCM configuration: %s", err.message);
+			return -dbus_error_to_errno(&err);
+		}
+
+	}
+
+#if BLUEALSA_HW_PARAMS_FIX
 	if (bluealsa_fix_hw_params(io, params) < 0)
 		debug2("Warning - unable to fix incorrect buffer size in hw parameters");
 #endif
 
 	snd_pcm_uframes_t period_size;
-	int ret;
-	if ((ret = snd_pcm_hw_params_get_period_size(params, &period_size, 0)) < 0)
+	if ((ret = snd_pcm_hw_params_get_period_size(params, &period_size, NULL)) < 0)
 		return ret;
 	snd_pcm_uframes_t buffer_size;
 	if ((ret = snd_pcm_hw_params_get_buffer_size(params, &buffer_size)) < 0)
@@ -560,12 +607,12 @@ static int bluealsa_hw_params(snd_pcm_ioplug_t *io, snd_pcm_hw_params_t *params)
 
 	pcm->frame_size = (snd_pcm_format_physical_width(io->format) * io->channels) / 8;
 
-	DBusError err = DBUS_ERROR_INIT;
-	if (!bluealsa_dbus_pcm_open(&pcm->dbus_ctx, pcm->ba_pcm.pcm_path,
+	if (!ba_dbus_pcm_open(&pcm->dbus_ctx, pcm->ba_pcm.pcm_path,
 				&pcm->ba_pcm_fd, &pcm->ba_pcm_ctrl_fd, &err)) {
 		debug2("Couldn't open PCM: %s", err.message);
+		ret = -dbus_error_to_errno(&err);
 		dbus_error_free(&err);
-		return -EBUSY;
+		return ret;
 	}
 
 	pcm->connected = true;
@@ -652,11 +699,20 @@ static int bluealsa_prepare(snd_pcm_ioplug_t *io) {
 	const snd_pcm_channel_area_t *areas = snd_pcm_ioplug_mmap_areas(io);
 	pcm->io_hw_buffer = (char *)areas->addr + areas->first / 8;
 
-	/* Indicate that our PCM is ready for IO, even though is is not 100%
-	 * true - the IO thread may not be running yet. Applications using
-	 * snd_pcm_sw_params_set_start_threshold() require the PCM to be usable
-	 * as soon as it has been prepared. */
-	eventfd_write(pcm->event_fd, 1);
+	if (io->stream == SND_PCM_STREAM_PLAYBACK) {
+		/* Indicate that our PCM is ready for IO, even though is is not 100%
+		 * true - the IO thread may not be running yet. Applications using
+		 * snd_pcm_sw_params_set_start_threshold() require the PCM to be usable
+		 * as soon as it has been prepared. */
+		if (pcm->io_avail_min < io->buffer_size)
+			eventfd_write(pcm->event_fd, 1);
+	}
+	else {
+		/* Make sure there is no poll event still pending (for example when
+		 * preparing after an overrun). */
+			eventfd_t event;
+			eventfd_read(pcm->event_fd, &event);
+	}
 
 	debug2("Prepared");
 	return 0;
@@ -675,42 +731,94 @@ static int bluealsa_drain(snd_pcm_ioplug_t *io) {
 	 * always either finishes in state SND_PCM_STATE_SETUP or returns an error.
 	 * It is not possible to finish in state SND_PCM_STATE_DRAINING and return
 	 * success; therefore is is impossible to correctly implement capture
-	 * drain logic. So for capture PCMs we do nothing and return success. */
+	 * drain logic. So for capture PCMs we do nothing and return success;
+	 * ioplug will stop the PCM. */
+	if (io->stream == SND_PCM_STREAM_CAPTURE)
+		return 0;
 
-	if (io->stream == SND_PCM_STREAM_PLAYBACK) {
+	/* We must ensure that all remaining frames in the ring buffer are flushed
+	 * to the FIFO by the I/O thread. It is possible that the client has called
+	 * snd_pcm_drain() without the start_threshold having been reached, or
+	 * while paused, so we must first ensure that the IO thread is running. */
+	if (bluealsa_start(io) < 0) {
+		/* Insufficient resources to start a new thread - so we have no choice
+		 * but to drop this stream. */
+		bluealsa_stop(io);
+		snd_pcm_ioplug_set_state(io, SND_PCM_STATE_SETUP);
+		return -EIO;
+	}
 
-		/* We must ensure that all remaining frames in the ring buffer are
-		 * flushed to the FIFO by the I/O thread. It is possible that the
-		 * client has called snd_pcm_drain() without the start_threshold
-		 * having been reached, or while paused, so we must first ensure that
-		 * the IO thread is running. */
-		if (bluealsa_start(io) < 0)
-			return 0;
+	/* For a non-blocking drain, we do not wait for the drain to complete. */
+	if (io->nonblock)
+		return -EAGAIN;
 
-		/* Block until the drain is complete. */
-		struct pollfd pfd = { pcm->event_fd, POLLIN, 0 };
-		while (bluealsa_pointer(io) >= 0 && io->state == SND_PCM_STATE_DRAINING) {
-			if (poll(&pfd, 1, -1) == -1) {
-				if (errno == EINTR)
+	struct pollfd pfd = { pcm->event_fd, POLLIN, 0 };
+	bool aborted = false;
+	int ret = 0;
+
+	snd_pcm_sframes_t hw_ptr;
+	while ((hw_ptr = bluealsa_pointer(io)) >= 0 && io->state == SND_PCM_STATE_DRAINING) {
+
+		snd_pcm_uframes_t avail = snd_pcm_ioplug_hw_avail(io, hw_ptr, io->appl_ptr);
+
+		/* If the buffer is empty then the local drain is complete. */
+		if (avail == 0)
+			break;
+
+		/* We set a timeout to ensure that the plugin cannot block forever in
+		 * case the server has stopped reading from the FIFO. Allow enough time
+		 * to drain the available frames as full periods, plus 100 ms:
+		 * e.g. one or less periods in buffer, allow 100ms + one period, more
+		 *      than one but not more than two, allow 100ms + two periods, etc.
+		 * If the wait is re-started after being interrupted by a signal then
+		 * we must re-calculate the maximum waiting time that remains. */
+		int timeout = 100 + (((avail - 1) / io->period_size) + 1) * io->period_size * 1000 / io->rate;
+
+		int nready = poll(&pfd, 1, timeout);
+		if (nready == -1) {
+			if (errno == EINTR) {
+				/* It is not well documented by ALSA, but if the application has
+				 * requested that the PCM should be aborted by a signal then the
+				 * ioplug nonblock flag is set to the special value 2. */
+				if (io->nonblock != 2)
 					continue;
-				break;
+				/* Application has aborted the drain. */
+				debug2("Drain aborted by signal");
+				aborted = true;
 			}
-			if (pfd.revents & POLLIN) {
-				eventfd_t event;
-				eventfd_read(pcm->event_fd, &event);
-				if (event & 0xDEAD0000)
-					break;
+			else {
+				debug2("Drain poll error: %s", strerror(errno));
+				bluealsa_stop(io);
+				snd_pcm_ioplug_set_state(io, SND_PCM_STATE_SETUP);
+				ret = -EIO;
 			}
+
+			break;
+		}
+		if (nready == 0) {
+			/* Timeout - do not wait any longer. */
+			SNDERR("Drain timed out: Possible Bluetooth transport failure");
+			bluealsa_stop(io);
+			io->state = SND_PCM_STATE_SETUP;
+			ret = -EIO;
+			break;
 		}
 
-		bluealsa_dbus_pcm_ctrl_send_drain(pcm->ba_pcm_ctrl_fd, NULL);
+		if (pfd.revents & POLLIN) {
+			eventfd_t event;
+			eventfd_read(pcm->event_fd, &event);
+		}
 
 	}
 
-	/* We cannot recover from an error here. By returning zero we ensure that
-	 * ioplug stops the pcm. Returning an error code would be interpreted by
-	 * ioplug as an incomplete drain and would it leave the pcm running. */
-	return 0;
+	if (io->state == SND_PCM_STATE_DRAINING && !aborted)
+		if (!ba_dbus_pcm_ctrl_send_drain(pcm->ba_pcm_ctrl_fd, NULL)) {
+			bluealsa_stop(io);
+			io->state = SND_PCM_STATE_SETUP;
+			ret = -EIO;
+		}
+
+	return ret;
 }
 
 /**
@@ -740,7 +848,7 @@ static snd_pcm_sframes_t bluealsa_calculate_delay(snd_pcm_ioplug_t *io) {
 	 * dispatching was done more than one second ago - this should prioritize
 	 * asynchronous dispatching in the poll_revents() callback. */
 	if (pcm->dbus_dispatch_ts.tv_sec + 1 < now.tv_sec) {
-		bluealsa_dbus_connection_dispatch(&pcm->dbus_ctx);
+		ba_dbus_connection_dispatch(&pcm->dbus_ctx);
 		gettimestamp(&pcm->dbus_dispatch_ts);
 	}
 
@@ -833,9 +941,9 @@ static int bluealsa_pause(snd_pcm_ioplug_t *io, int enable) {
 		return -ENODEV;
 	}
 
-	if (!bluealsa_dbus_pcm_ctrl_send(pcm->ba_pcm_ctrl_fd,
-				enable ? "Pause" : "Resume", NULL))
-		return -errno;
+	if (!ba_dbus_pcm_ctrl_send(pcm->ba_pcm_ctrl_fd,
+				enable ? "Pause" : "Resume", 200, NULL))
+		return -EIO;
 
 	if (enable == 0)
 		pthread_kill(pcm->io_thread, SIGIO);
@@ -903,7 +1011,7 @@ static int bluealsa_poll_descriptors_count(snd_pcm_ioplug_t *io) {
 	struct bluealsa_pcm *pcm = io->private_data;
 
 	nfds_t dbus_nfds = 0;
-	bluealsa_dbus_connection_poll_fds(&pcm->dbus_ctx, NULL, &dbus_nfds);
+	ba_dbus_connection_poll_fds(&pcm->dbus_ctx, NULL, &dbus_nfds);
 
 	return 1 + dbus_nfds;
 }
@@ -916,7 +1024,7 @@ static int bluealsa_poll_descriptors(snd_pcm_ioplug_t *io, struct pollfd *pfd,
 		return -EINVAL;
 
 	nfds_t dbus_nfds = nfds - 1;
-	if (!bluealsa_dbus_connection_poll_fds(&pcm->dbus_ctx, &pfd[1], &dbus_nfds))
+	if (!ba_dbus_connection_poll_fds(&pcm->dbus_ctx, &pfd[1], &dbus_nfds))
 		return -EINVAL;
 
 	/* PCM plug-in relies on our internal event file descriptor. */
@@ -936,7 +1044,7 @@ static int bluealsa_poll_revents(snd_pcm_ioplug_t *io, struct pollfd *pfd,
 	if (nfds < 1)
 		return -EINVAL;
 
-	bluealsa_dbus_connection_poll_dispatch(&pcm->dbus_ctx, &pfd[1], nfds - 1);
+	ba_dbus_connection_poll_dispatch(&pcm->dbus_ctx, &pfd[1], nfds - 1);
 	while (dbus_connection_dispatch(pcm->dbus_ctx.conn) == DBUS_DISPATCH_DATA_REMAINS)
 		continue;
 	gettimestamp(&pcm->dbus_dispatch_ts);
@@ -953,7 +1061,8 @@ static int bluealsa_poll_revents(snd_pcm_ioplug_t *io, struct pollfd *pfd,
 			goto fail;
 
 		/* This call synchronizes the ring buffer pointers and updates the
-		 * ioplug state. */
+		 * ioplug state. For non-blocking drains it also causes ioplug to drop
+		 * the stream when the buffer is empty. */
 		snd_pcm_sframes_t avail = snd_pcm_avail(io->pcm);
 
 		/* ALSA expects that the event will match stream direction, e.g.
@@ -966,8 +1075,13 @@ static int bluealsa_poll_revents(snd_pcm_ioplug_t *io, struct pollfd *pfd,
 
 		switch (io->state) {
 			case SND_PCM_STATE_SETUP:
+				/* To support non-blocking drain we must report a POLLOUT event
+				 * for playback PCMs here, because the above call to
+				 * snd_pcm_avail() may have changed the state to
+				 * SND_PCM_STATE_SETUP. */
+				if (io->stream == SND_PCM_STREAM_CAPTURE)
+					*revents = 0;
 				ready = false;
-				*revents = 0;
 				break;
 			case SND_PCM_STATE_PREPARED:
 				/* capture poll should block forever */
@@ -982,6 +1096,15 @@ static int bluealsa_poll_revents(snd_pcm_ioplug_t *io, struct pollfd *pfd,
 					*revents = 0;
 				}
 				break;
+			case SND_PCM_STATE_DRAINING:
+				/* BlueALSA does not drain capture PCMs. So this state only
+				 * occurs with playback PCMs. Do not wake the application until
+				 * the buffer is empty. */
+				if ((snd_pcm_uframes_t)avail < io->buffer_size) {
+					ready = false;
+					*revents = 0;
+				}
+				break;
 			case SND_PCM_STATE_XRUN:
 			case SND_PCM_STATE_PAUSED:
 			case SND_PCM_STATE_SUSPENDED:
@@ -1035,8 +1158,7 @@ static int str2bdaddr(const char *str, bdaddr_t *ba) {
 				&x[5], &x[4], &x[3], &x[2], &x[1], &x[0]) != 6)
 		return -1;
 
-	size_t i;
-	for (i = 0; i < 6; i++)
+	for (size_t i = 0; i < 6; i++)
 		ba->b[i] = x[i];
 
 	return 0;
@@ -1051,6 +1173,41 @@ static int str2profile(const char *str) {
 	return 0;
 }
 
+/**
+ * Extract codec name and configuration from the codec string. */
+static int str2codec(const char *codec, char *name, size_t name_size,
+		uint8_t *config, size_t config_size, size_t *config_len) {
+
+	size_t name_len = strlen(codec);
+
+	char *delim;
+	/* Check for the delimiter which separates codec name and configuration. */
+	if ((delim = strchr(codec, ':')) != NULL) {
+
+		name_len = delim - codec;
+		delim++;
+
+		size_t config_hex_len;
+		if ((config_hex_len = strlen(delim)) > config_size * 2)
+			return -1;
+
+		ssize_t len;
+		if ((len = hex2bin(delim, config, config_hex_len)) == -1)
+			return -1;
+
+		*config_len = len;
+
+	}
+
+	if (name_len >= name_size)
+		return -1;
+
+	strncpy(name, codec, name_len);
+	name[name_len] = '\0';
+
+	return 0;
+}
+
 /**
  * Convert volume string to volume level and mute state.
  *
@@ -1121,7 +1278,7 @@ static DBusHandlerResult bluealsa_dbus_msg_filter(DBusConnection *conn,
 	dbus_message_iter_next(&iter);
 
 	if (strcmp(updated_interface, BLUEALSA_INTERFACE_PCM) == 0)
-		bluealsa_dbus_message_iter_get_pcm_props(&iter, NULL, &pcm->ba_pcm);
+		dbus_message_iter_get_ba_pcm_props(&iter, NULL, &pcm->ba_pcm);
 
 	return DBUS_HANDLER_RESULT_HANDLED;
 }
@@ -1138,6 +1295,13 @@ static int bluealsa_set_hw_constraint(struct bluealsa_pcm *pcm) {
 
 	debug2("Setting constraints");
 
+	const struct ba_pcm_codec *codec = &pcm->ba_pcm.codec;
+	for (size_t i = 0; i < pcm->ba_pcm_codecs.codecs_len; i++)
+		if (strcmp(pcm->ba_pcm_codecs.codecs[i].name, codec->name) == 0) {
+			codec = &pcm->ba_pcm_codecs.codecs[i];
+			break;
+		}
+
 	if ((err = snd_pcm_ioplug_set_param_list(io, SND_PCM_IOPLUG_HW_ACCESS,
 					ARRAYSIZE(accesses), accesses)) < 0)
 		return err;
@@ -1163,61 +1327,30 @@ static int bluealsa_set_hw_constraint(struct bluealsa_pcm *pcm) {
 					min_p, 1024 * 1024)) < 0)
 		return err;
 
-	if ((err = snd_pcm_ioplug_set_param_minmax(io, SND_PCM_IOPLUG_HW_BUFFER_BYTES,
-					2 * min_p, 2 * 1024 * 1024)) < 0)
+	unsigned int list[ARRAYSIZE(codec->sampling)];
+	unsigned int n;
+
+	/* Populate the list of supported channels and sampling rates. For codecs
+	 * with fixed configuration, the list will contain only one element. For
+	 * other codecs, the list might contain all supported configurations. */
+
+	n = 0;
+	for (size_t i = 0; i < ARRAYSIZE(codec->channels) && codec->channels[i] != 0; i++)
+		/* TODO: Remove this when BlueALSA will support channels mapping. */
+		if (codec->channels[i] <= 2)
+			list[n++] = codec->channels[i];
+	if ((err = snd_pcm_ioplug_set_param_list(io, SND_PCM_IOPLUG_HW_CHANNELS, n, list)) < 0)
 		return err;
 
-	if ((err = snd_pcm_ioplug_set_param_minmax(io, SND_PCM_IOPLUG_HW_CHANNELS,
-					pcm->ba_pcm.channels, pcm->ba_pcm.channels)) < 0)
-		return err;
-
-	if ((err = snd_pcm_ioplug_set_param_minmax(io, SND_PCM_IOPLUG_HW_RATE,
-					pcm->ba_pcm.sampling, pcm->ba_pcm.sampling)) < 0)
+	n = 0;
+	for (size_t i = 0; i < ARRAYSIZE(codec->sampling) && codec->sampling[i] != 0; i++)
+		list[n++] = codec->sampling[i];
+	if ((err = snd_pcm_ioplug_set_param_list(io, SND_PCM_IOPLUG_HW_RATE, n, list)) < 0)
 		return err;
 
 	return 0;
 }
 
-static bool bluealsa_select_pcm_codec(struct bluealsa_pcm *pcm, const char *codec, DBusError *err) {
-
-	uint8_t config[64];
-	ssize_t config_len = 0;
-	bool ret = false;
-
-	char *codec_name;
-	if ((codec_name = strdup(codec)) == NULL) {
-		dbus_set_error(err, DBUS_ERROR_NO_MEMORY, NULL);
-		return false;
-	}
-
-	char *config_hex;
-	/* split the given string into name and configuration components */
-	if ((config_hex = strchr(codec_name, ':')) != NULL) {
-		*config_hex++ = '\0';
-
-		size_t config_hex_len;
-		if ((config_hex_len = strlen(config_hex)) > sizeof(config) * 2) {
-			dbus_set_error(err, DBUS_ERROR_FAILED, "Invalid codec configuration: %s", config_hex);
-			goto fail;
-		}
-		if ((config_len = hex2bin(config_hex, config, config_hex_len)) == -1) {
-			dbus_set_error(err, DBUS_ERROR_FAILED, "%s", strerror(errno));
-			goto fail;
-		}
-
-	}
-
-	if (!bluealsa_dbus_pcm_select_codec(&pcm->dbus_ctx, pcm->ba_pcm.pcm_path,
-				bluealsa_dbus_pcm_get_codec_canonical_name(codec_name), config, config_len, err))
-		goto fail;
-
-	ret = true;
-
-fail:
-	free(codec_name);
-	return ret;
-}
-
 static bool bluealsa_update_pcm_volume(struct bluealsa_pcm *pcm,
 		int volume, int mute, DBusError *err) {
 	uint16_t old = pcm->ba_pcm.volume.raw;
@@ -1235,7 +1368,7 @@ static bool bluealsa_update_pcm_volume(struct bluealsa_pcm *pcm,
 	}
 	if (pcm->ba_pcm.volume.raw == old)
 		return true;
-	return bluealsa_dbus_pcm_update(&pcm->dbus_ctx, &pcm->ba_pcm, BLUEALSA_PCM_VOLUME, err);
+	return ba_dbus_pcm_update(&pcm->dbus_ctx, &pcm->ba_pcm, BLUEALSA_PCM_VOLUME, err);
 }
 
 static bool bluealsa_update_pcm_softvol(struct bluealsa_pcm *pcm,
@@ -1243,7 +1376,7 @@ static bool bluealsa_update_pcm_softvol(struct bluealsa_pcm *pcm,
 	if (softvol < 0 || !!softvol == pcm->ba_pcm.soft_volume)
 		return true;
 	pcm->ba_pcm.soft_volume = !!softvol;
-	return bluealsa_dbus_pcm_update(&pcm->dbus_ctx, &pcm->ba_pcm, BLUEALSA_PCM_SOFT_VOLUME, err);
+	return ba_dbus_pcm_update(&pcm->dbus_ctx, &pcm->ba_pcm, BLUEALSA_PCM_SOFT_VOLUME, err);
 }
 
 SND_PCM_PLUGIN_DEFINE_FUNC(bluealsa) {
@@ -1344,6 +1477,15 @@ SND_PCM_PLUGIN_DEFINE_FUNC(bluealsa) {
 		return -EINVAL;
 	}
 
+	char codec_name[32] = "";
+	uint8_t codec_config[sizeof(pcm->ba_pcm_codec_config)];
+	size_t codec_config_len = 0;
+	if (codec != NULL && str2codec(codec, codec_name, sizeof(codec_name),
+				codec_config, sizeof(codec_config), &codec_config_len) == -1) {
+		SNDERR("Invalid codec: %s", codec);
+		return -EINVAL;
+	}
+
 	int pcm_mute = -1;
 	int pcm_volume = -1;
 	if (volume != NULL && str2volume(volume, &pcm_volume, &pcm_mute) != 0) {
@@ -1360,6 +1502,16 @@ SND_PCM_PLUGIN_DEFINE_FUNC(bluealsa) {
 	if ((pcm = calloc(1, sizeof(*pcm))) == NULL)
 		return -ENOMEM;
 
+	pcm->io.version = SND_PCM_IOPLUG_VERSION;
+	pcm->io.name = "BlueALSA";
+	pcm->io.flags = SND_PCM_IOPLUG_FLAG_LISTED;
+#ifdef SND_PCM_IOPLUG_FLAG_BOUNDARY_WA
+	pcm->io.flags |= SND_PCM_IOPLUG_FLAG_BOUNDARY_WA;
+#endif
+	pcm->io.mmap_rw = 1;
+	pcm->io.callback = &bluealsa_callback;
+	pcm->io.private_data = pcm;
+
 	pcm->event_fd = -1;
 	pcm->ba_pcm_fd = -1;
 	pcm->ba_pcm_ctrl_fd = -1;
@@ -1371,9 +1523,9 @@ SND_PCM_PLUGIN_DEFINE_FUNC(bluealsa) {
 	dbus_threads_init_default();
 
 	DBusError err = DBUS_ERROR_INIT;
-	if (bluealsa_dbus_connection_ctx_init(&pcm->dbus_ctx, service, &err) != TRUE) {
+	if (ba_dbus_connection_ctx_init(&pcm->dbus_ctx, service, &err) != TRUE) {
 		SNDERR("Couldn't initialize D-Bus context: %s", err.message);
-		ret = -ENOMEM;
+		ret = -dbus_error_to_errno(&err);
 		goto fail;
 	}
 
@@ -1384,49 +1536,49 @@ SND_PCM_PLUGIN_DEFINE_FUNC(bluealsa) {
 	}
 
 	debug("Getting BlueALSA PCM: %s %s %s", snd_pcm_stream_name(stream), device, profile);
-	if (!bluealsa_dbus_get_pcm(&pcm->dbus_ctx, &ba_addr, ba_profile,
+	if (!ba_dbus_pcm_get(&pcm->dbus_ctx, &ba_addr, ba_profile,
 				stream == SND_PCM_STREAM_PLAYBACK ? BA_PCM_MODE_SINK : BA_PCM_MODE_SOURCE,
 				&pcm->ba_pcm, &err)) {
 		SNDERR("Couldn't get BlueALSA PCM: %s", err.message);
-		ret = -ENODEV;
+		ret = -dbus_error_to_errno(&err);
 		goto fail;
 	}
 
 	/* Subscribe for properties-changed signals but for the opened PCM only. */
-	bluealsa_dbus_connection_signal_match_add(&pcm->dbus_ctx, pcm->dbus_ctx.ba_service,
+	ba_dbus_connection_signal_match_add(&pcm->dbus_ctx, pcm->dbus_ctx.ba_service,
 			pcm->ba_pcm.pcm_path, DBUS_INTERFACE_PROPERTIES, "PropertiesChanged",
 			"arg0='"BLUEALSA_INTERFACE_PCM"'");
 
-	if ((pcm->event_fd = eventfd(0, EFD_CLOEXEC)) == -1) {
+	if ((pcm->event_fd = eventfd(0, EFD_CLOEXEC | EFD_NONBLOCK)) == -1) {
 		ret = -errno;
 		goto fail;
 	}
 
-	pcm->io.version = SND_PCM_IOPLUG_VERSION;
-	pcm->io.name = "BlueALSA";
-	pcm->io.flags = SND_PCM_IOPLUG_FLAG_LISTED;
-#ifdef SND_PCM_IOPLUG_FLAG_BOUNDARY_WA
-	pcm->io.flags |= SND_PCM_IOPLUG_FLAG_BOUNDARY_WA;
-#endif
-	pcm->io.mmap_rw = 1;
-	pcm->io.callback = &bluealsa_callback;
-	pcm->io.private_data = pcm;
+	if (codec_name[0] != '\0') {
+		/* If the codec was given, change it now, so we can get the correct
+		 * sampling rate and channels for HW constraints. */
+		const char *canonical = ba_dbus_pcm_codec_get_canonical_name(codec_name);
+		const bool name_changed = strcmp(canonical, pcm->ba_pcm.codec.name) != 0;
+		if (name_changed && !ba_dbus_pcm_select_codec(&pcm->dbus_ctx, pcm->ba_pcm.pcm_path,
+					canonical, NULL, 0, 0, 0, BA_PCM_SELECT_CODEC_FLAG_NONE, &err)) {
+			SNDERR("Couldn't select BlueALSA PCM codec: %s", err.message);
+			dbus_error_free(&err);
+		}
+		else {
+
+			memcpy(pcm->ba_pcm_codec_config, codec_config, codec_config_len);
+			pcm->ba_pcm_codec_config_len = codec_config_len;
 
-	if (codec != NULL && codec[0] != '\0') {
-		if (bluealsa_select_pcm_codec(pcm, codec, &err)) {
 			/* Changing the codec may change the audio format, sampling rate and/or
 			 * channels. We need to refresh our cache of PCM properties. */
-			if (!bluealsa_dbus_get_pcm(&pcm->dbus_ctx, &ba_addr, ba_profile,
+			if (name_changed && !ba_dbus_pcm_get(&pcm->dbus_ctx, &ba_addr, ba_profile,
 						stream == SND_PCM_STREAM_PLAYBACK ? BA_PCM_MODE_SINK : BA_PCM_MODE_SOURCE,
 						&pcm->ba_pcm, &err)) {
 				SNDERR("Couldn't get BlueALSA PCM: %s", err.message);
-				ret = -ENODEV;
+				ret = -dbus_error_to_errno(&err);
 				goto fail;
 			}
-		}
-		else {
-			SNDERR("Couldn't select BlueALSA PCM codec: %s", err.message);
-			dbus_error_free(&err);
+
 		}
 	}
 
@@ -1448,8 +1600,14 @@ SND_PCM_PLUGIN_DEFINE_FUNC(bluealsa) {
 	if ((ret = snd_pcm_ioplug_create(&pcm->io, name, stream, mode)) < 0)
 		goto fail;
 
-	if ((ret = bluealsa_set_hw_constraint(pcm)) < 0)
+	if (!ba_dbus_pcm_codecs_get(&pcm->dbus_ctx, pcm->ba_pcm.pcm_path,
+				&pcm->ba_pcm_codecs, &err))
+		SNDERR("Couldn't get BlueALSA PCM codecs: %s", err.message);
+
+	if ((ret = bluealsa_set_hw_constraint(pcm)) < 0) {
+		snd_pcm_ioplug_delete(&pcm->io);
 		goto fail;
+	}
 
 	if (!bluealsa_update_pcm_softvol(pcm, pcm_softvol, &err)) {
 		SNDERR("Couldn't set BlueALSA PCM soft-volume: %s", err.message);
@@ -1465,13 +1623,8 @@ SND_PCM_PLUGIN_DEFINE_FUNC(bluealsa) {
 	return 0;
 
 fail:
-	bluealsa_dbus_connection_ctx_free(&pcm->dbus_ctx);
+	bluealsa_close(&pcm->io);
 	dbus_error_free(&err);
-	if (pcm->event_fd != -1)
-		close(pcm->event_fd);
-	pthread_mutex_destroy(&pcm->mutex);
-	pthread_cond_destroy(&pcm->pause_cond);
-	free(pcm);
 	return ret;
 }
 
