@@ -1,6 +1,6 @@
 /*
  * BlueALSA - ba-rfcomm.c
- * Copyright (c) 2016-2023 Arkadiusz Bokowy
+ * Copyright (c) 2016-2024 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -9,7 +9,10 @@
  */
 
 #include "ba-rfcomm.h"
-/* IWYU pragma: no_include "config.h" */
+
+#if HAVE_CONFIG_H
+# include <config.h>
+#endif
 
 #include <errno.h>
 #include <poll.h>
@@ -25,9 +28,10 @@
 #include <glib.h>
 
 #include "ba-adapter.h"
+#include "ba-config.h"
 #include "ba-device.h"
 #include "ba-transport.h"
-#include "bluealsa-config.h"
+#include "ba-transport-pcm.h"
 #include "bluealsa-dbus.h"
 #include "bluez.h"
 #include "shared/defs.h"
@@ -224,10 +228,9 @@ static int rfcomm_handler_cind_resp_get_cb(struct ba_rfcomm *r, const struct bt_
 
 	struct ba_device * const d = r->sco->d;
 	char *tmp = at->value;
-	size_t i;
 
 	/* parse response for the +CIND GET command */
-	for (i = 0; i < ARRAYSIZE(r->hfp_ind_map); i++) {
+	for (size_t i = 0; i < ARRAYSIZE(r->hfp_ind_map); i++) {
 		r->hfp_ind[r->hfp_ind_map[i]] = atoi(tmp);
 		if (r->hfp_ind_map[i] == HFP_IND_BATTCHG) {
 			d->battery.charge = atoi(tmp) * 100 / 5;
@@ -359,12 +362,16 @@ static int rfcomm_handler_brsf_set_cb(struct ba_rfcomm *r, const struct bt_at *a
 	if (!(r->ag_features & HFP_AG_FEAT_CODEC)) {
 		/* Assume that mandatory codec is supported. */
 		r->hf_codecs.cvsd = true;
-#if ENABLE_MSBC
 		/* If codec selection is supported assume that
-		 * mSBC is supported as well. */
-		if (r->hf_features & HFP_HF_FEAT_CODEC)
+		 * mSBC and/or LC3-SWB are supported as well. */
+		if (r->hf_features & HFP_HF_FEAT_CODEC) {
+#if ENABLE_MSBC
 			r->hf_codecs.msbc = true;
 #endif
+#if ENABLE_LC3_SWB
+			r->hf_codecs.lc3_swb = true;
+#endif
+		}
 	}
 
 	sprintf(tmp, "%u", r->ag_features);
@@ -398,12 +405,16 @@ static int rfcomm_handler_brsf_resp_cb(struct ba_rfcomm *r, const struct bt_at *
 	 * we can assume that AG supports it. */
 	r->ag_codecs.cvsd = true;
 
+	/* If codec selection is supported in the AG, we can assume
+	 * that mSBC and/or LC3-SWB are supported as well. */
+	if (r->ag_features & HFP_AG_FEAT_CODEC) {
 #if ENABLE_MSBC
-	/* If codec selection is supported in the AG, we
-	 * can assume that mSBC is supported as well. */
-	if (r->ag_features & HFP_AG_FEAT_CODEC)
 		r->ag_codecs.msbc = true;
 #endif
+#if ENABLE_LC3_SWB
+		r->ag_codecs.lc3_swb = true;
+#endif
+	}
 
 	if (r->state < HFP_SLC_BRSF_SET)
 		rfcomm_set_hfp_state(r, HFP_SLC_BRSF_SET);
@@ -428,7 +439,7 @@ static int rfcomm_handler_nrec_set_cb(struct ba_rfcomm *r, const struct bt_at *a
 static int rfcomm_handler_vgm_set_cb(struct ba_rfcomm *r, const struct bt_at *at) {
 
 	struct ba_transport * const t_sco = r->sco;
-	struct ba_transport_pcm *pcm = &t_sco->sco.mic_pcm;
+	struct ba_transport_pcm *pcm = &t_sco->sco.pcm_mic;
 	const int fd = r->fd;
 
 	/* skip update in case of software volume */
@@ -454,7 +465,7 @@ static int rfcomm_handler_vgm_set_cb(struct ba_rfcomm *r, const struct bt_at *at
 static int rfcomm_handler_vgm_resp_cb(struct ba_rfcomm *r, const struct bt_at *at) {
 
 	struct ba_transport * const t_sco = r->sco;
-	struct ba_transport_pcm *pcm = &t_sco->sco.mic_pcm;
+	struct ba_transport_pcm *pcm = &t_sco->sco.pcm_mic;
 
 	r->gain_mic = atoi(at->value);
 	int level = ba_transport_pcm_volume_range_to_level(r->gain_mic, HFP_VOLUME_GAIN_MAX);
@@ -473,7 +484,7 @@ static int rfcomm_handler_vgm_resp_cb(struct ba_rfcomm *r, const struct bt_at *a
 static int rfcomm_handler_vgs_set_cb(struct ba_rfcomm *r, const struct bt_at *at) {
 
 	struct ba_transport * const t_sco = r->sco;
-	struct ba_transport_pcm *pcm = &t_sco->sco.spk_pcm;
+	struct ba_transport_pcm *pcm = &t_sco->sco.pcm_spk;
 	const int fd = r->fd;
 
 	/* skip update in case of software volume */
@@ -499,7 +510,7 @@ static int rfcomm_handler_vgs_set_cb(struct ba_rfcomm *r, const struct bt_at *at
 static int rfcomm_handler_vgs_resp_cb(struct ba_rfcomm *r, const struct bt_at *at) {
 
 	struct ba_transport * const t_sco = r->sco;
-	struct ba_transport_pcm *pcm = &t_sco->sco.spk_pcm;
+	struct ba_transport_pcm *pcm = &t_sco->sco.pcm_spk;
 
 	r->gain_spk = atoi(at->value);
 	int level = ba_transport_pcm_volume_range_to_level(r->gain_spk, HFP_VOLUME_GAIN_MAX);
@@ -525,7 +536,7 @@ static int rfcomm_handler_btrh_get_cb(struct ba_rfcomm *r, const struct bt_at *a
 	return 0;
 }
 
-#if ENABLE_MSBC
+#if ENABLE_HFP_CODEC_SELECTION
 static int rfcomm_hfp_setup_codec_connection(struct ba_rfcomm *r);
 #endif
 
@@ -534,7 +545,7 @@ static int rfcomm_hfp_setup_codec_connection(struct ba_rfcomm *r);
 static int rfcomm_handler_bcc_cmd_cb(struct ba_rfcomm *r, const struct bt_at *at) {
 	(void)at;
 	const int fd = r->fd;
-#if ENABLE_MSBC
+#if ENABLE_HFP_CODEC_SELECTION
 	if (rfcomm_write_at(fd, AT_TYPE_RESP, NULL, "OK") == -1)
 		return -1;
 	if (rfcomm_hfp_setup_codec_connection(r) == -1)
@@ -554,9 +565,9 @@ static int rfcomm_handler_bcs_set_cb(struct ba_rfcomm *r, const struct bt_at *at
 	const int fd = r->fd;
 	int rv;
 
-	int codec;
-	if ((codec = atoi(at->value)) != r->codec) {
-		warn("Codec not acknowledged: %s != %d", at->value, r->codec);
+	uint8_t codec_id;
+	if ((codec_id = atoi(at->value)) != r->codec_id) {
+		warn("Codec not acknowledged: %s != %u", at->value, r->codec_id);
 		rv = rfcomm_write_at(fd, AT_TYPE_RESP, NULL, "ERROR");
 		goto final;
 	}
@@ -566,7 +577,7 @@ static int rfcomm_handler_bcs_set_cb(struct ba_rfcomm *r, const struct bt_at *at
 
 	/* Codec negotiation process is complete. Update transport and
 	 * notify connected clients, that transport has been changed. */
-	ba_transport_set_codec(t_sco, codec);
+	ba_transport_set_codec(t_sco, codec_id);
 
 final:
 	rfcomm_finalize_codec_selection(r);
@@ -581,7 +592,7 @@ static int rfcomm_handler_resp_bcs_ok_cb(struct ba_rfcomm *r, const struct bt_at
 		return -1;
 
 	if (!r->handler_resp_ok_success) {
-		warn("Codec selection not finalized: %d", r->codec);
+		warn("Codec selection not finalized: %u", r->codec_id);
 		ba_transport_set_codec(t_sco, HFP_CODEC_UNDEFINED);
 		rfcomm_finalize_codec_selection(r);
 	}
@@ -598,18 +609,30 @@ static int rfcomm_handler_bcs_resp_cb(struct ba_rfcomm *r, const struct bt_at *a
 	static const struct ba_rfcomm_handler handler_unsupported = {
 		AT_TYPE_RESP, "", rfcomm_handler_resp_ok_cb };
 
-	const int fd = r->fd;
-	const int codec = atoi(at->value);
-	bool codec_supported = false;
-
-	if (r->hf_codecs.cvsd && codec == HFP_CODEC_CVSD)
-		codec_supported = true;
-#ifdef ENABLE_MSBC
-	if (r->hf_codecs.msbc && codec == HFP_CODEC_MSBC)
-		codec_supported = true;
+	const struct {
+		uint8_t codec_id;
+		bool is_supported;
+	} codecs[] = {
+		{ HFP_CODEC_CVSD, r->hf_codecs.cvsd },
+#if ENABLE_MSBC
+		{ HFP_CODEC_MSBC, r->hf_codecs.msbc },
 #endif
+#if ENABLE_LC3_SWB
+		{ HFP_CODEC_LC3_SWB, r->hf_codecs.lc3_swb },
+#endif
+	};
 
-	if (!codec_supported) {
+	const int fd = r->fd;
+	const uint8_t codec_id = atoi(at->value);
+
+	bool is_codec_supported = false;
+	for (size_t i = 0; i < ARRAYSIZE(codecs); i++)
+		if (codecs[i].codec_id == codec_id && codecs[i].is_supported) {
+			is_codec_supported = true;
+			break;
+		}
+
+	if (!is_codec_supported) {
 		/* If the requested codec is not supported, we must reply with the
 		 * list of codecs that we do support. */
 		if (rfcomm_write_at(fd, AT_TYPE_CMD_SET, "+BAC", r->hf_bac_bcs_string) == -1)
@@ -618,7 +641,7 @@ static int rfcomm_handler_bcs_resp_cb(struct ba_rfcomm *r, const struct bt_at *a
 		return 0;
 	}
 
-	r->codec = codec;
+	r->codec_id = codec_id;
 	if (rfcomm_write_at(fd, AT_TYPE_CMD_SET, "+BCS", at->value) == -1)
 		return -1;
 	r->handler = &handler_supported;
@@ -629,7 +652,7 @@ static int rfcomm_handler_bcs_resp_cb(struct ba_rfcomm *r, const struct bt_at *a
 	 * here and notify connected clients, that the transport has been
 	 * changed. Note, that this event might be emitted for an active
 	 * transport - codec switching initiated by Audio Gateway. */
-	ba_transport_set_codec(r->sco, r->codec);
+	ba_transport_set_codec(r->sco, r->codec_id);
 	rfcomm_finalize_codec_selection(r);
 
 	return 0;
@@ -657,6 +680,11 @@ static int rfcomm_handler_bac_set_cb(struct ba_rfcomm *r, const struct bt_at *at
 		case HFP_CODEC_MSBC:
 			r->hf_codecs.msbc = true;
 			break;
+#endif
+#if ENABLE_LC3_SWB
+		case HFP_CODEC_LC3_SWB:
+			r->hf_codecs.lc3_swb = true;
+			break;
 #endif
 		}
 	} while ((tmp = strchr(tmp, ',')) != NULL);
@@ -684,7 +712,7 @@ static int rfcomm_handler_android_set_xhsmicmute(struct ba_rfcomm *r, char *valu
 		return errno = EINVAL, -1;
 
 	struct ba_transport * const t_sco = r->sco;
-	struct ba_transport_pcm *pcm = &t_sco->sco.mic_pcm;
+	struct ba_transport_pcm *pcm = &t_sco->sco.pcm_mic;
 	const bool muted = value[0] == '0' ? false : true;
 	const int fd = r->fd;
 
@@ -756,8 +784,7 @@ static int rfcomm_handler_android_set_cb(struct ba_rfcomm *r, const struct bt_at
 	char *value = at->value;
 	char *name = strsep(&value, sep);
 
-	size_t i;
-	for (i = 0; i < ARRAYSIZE(handlers); i++)
+	for (size_t i = 0; i < ARRAYSIZE(handlers); i++)
 		if (strcmp(name, handlers[i].name) == 0) {
 			int rv = handlers[i].cb(r, value);
 			if (rv == -1 && errno == EINVAL)
@@ -928,9 +955,7 @@ static ba_rfcomm_callback *rfcomm_get_callback(const struct bt_at *at) {
 		&rfcomm_handler_xapl_resp,
 	};
 
-	size_t i;
-
-	for (i = 0; i < ARRAYSIZE(handlers); i++) {
+	for (size_t i = 0; i < ARRAYSIZE(handlers); i++) {
 		if (handlers[i]->type != at->type)
 			continue;
 		if (strcmp(handlers[i]->command, at->command) != 0)
@@ -957,11 +982,11 @@ static enum ba_rfcomm_signal rfcomm_recv_signal(struct ba_rfcomm *r) {
 	return BA_RFCOMM_SIGNAL_PING;
 }
 
-#if ENABLE_MSBC
+#if ENABLE_HFP_CODEC_SELECTION
 
 /**
  * Set HFP codec for the given Service Level Connection. */
-static int rfcomm_hfp_set_codec(struct ba_rfcomm *r, uint16_t codec) {
+static int rfcomm_hfp_set_codec(struct ba_rfcomm *r, uint8_t codec_id) {
 
 	struct ba_transport * const t_sco = r->sco;
 	const int fd = r->fd;
@@ -969,7 +994,7 @@ static int rfcomm_hfp_set_codec(struct ba_rfcomm *r, uint16_t codec) {
 
 	debug("RFCOMM: %s setting codec: %s",
 			ba_transport_debug_name(t_sco),
-			hfp_codec_id_to_string(codec));
+			hfp_codec_id_to_string(codec_id));
 
 	/* SLC is required for codec connection */
 	if (r->state != HFP_SLC_CONNECTED)
@@ -980,11 +1005,11 @@ static int rfcomm_hfp_set_codec(struct ba_rfcomm *r, uint16_t codec) {
 		goto fail;
 
 	char tmp[16];
-	sprintf(tmp, "%d", codec);
+	sprintf(tmp, "%u", codec_id);
 	if ((rv = rfcomm_write_at(fd, AT_TYPE_RESP, "+BCS", tmp)) == -1)
 		goto fail;
 
-	r->codec = codec;
+	r->codec_id = codec_id;
 	r->handler = &rfcomm_handler_bcs_set;
 	return 0;
 
@@ -997,11 +1022,16 @@ fail:
  * Try to setup HFP codec connection. */
 static int rfcomm_hfp_setup_codec_connection(struct ba_rfcomm *r) {
 
-	struct {
-		uint16_t codec_id;
+	const struct {
+		uint8_t codec_id;
 		bool is_supported;
 	} codecs[] = {
+#if ENABLE_LC3_SWB
+		{ HFP_CODEC_LC3_SWB, r->ag_codecs.lc3_swb && r->hf_codecs.lc3_swb },
+#endif
+#if ENABLE_MSBC
 		{ HFP_CODEC_MSBC, r->ag_codecs.msbc && r->hf_codecs.msbc },
+#endif
 		{ HFP_CODEC_CVSD, r->ag_codecs.cvsd && r->hf_codecs.cvsd },
 	};
 
@@ -1075,7 +1105,7 @@ static int rfcomm_notify_battery_level_change(struct ba_rfcomm *r) {
 static int rfcomm_notify_volume_change_mic(struct ba_rfcomm *r, bool force) {
 
 	struct ba_transport * const t_sco = r->sco;
-	struct ba_transport_pcm *pcm = &t_sco->sco.mic_pcm;
+	struct ba_transport_pcm *pcm = &t_sco->sco.pcm_mic;
 	const int fd = r->fd;
 	char tmp[24];
 
@@ -1107,7 +1137,7 @@ static int rfcomm_notify_volume_change_mic(struct ba_rfcomm *r, bool force) {
 static int rfcomm_notify_volume_change_spk(struct ba_rfcomm *r, bool force) {
 
 	struct ba_transport * const t_sco = r->sco;
-	struct ba_transport_pcm *pcm = &t_sco->sco.spk_pcm;
+	struct ba_transport_pcm *pcm = &t_sco->sco.pcm_spk;
 	const int fd = r->fd;
 	char tmp[24];
 
@@ -1176,7 +1206,7 @@ static void *rfcomm_thread(struct ba_rfcomm *r) {
 	pthread_cleanup_push(PTHREAD_CLEANUP(rfcomm_thread_cleanup), r);
 
 	sigset_t sigset;
-	/* See the ba_transport_thread_create() function for information
+	/* See the ba_transport_pcm_start() function for information
 	 * why we have to mask all signals. */
 	sigfillset(&sigset);
 	pthread_sigmask(SIG_SETMASK, &sigset, NULL);
@@ -1287,9 +1317,9 @@ static void *rfcomm_thread(struct ba_rfcomm *r) {
 					/* If codec was selected during the SLC establishment,
 					 * notify BlueALSA D-Bus clients about the change. */
 					if (ba_transport_get_codec(t_sco) != HFP_CODEC_UNDEFINED) {
-						bluealsa_dbus_pcm_update(&t_sco->sco.spk_pcm,
+						bluealsa_dbus_pcm_update(&t_sco->sco.pcm_spk,
 								BA_DBUS_PCM_UPDATE_SAMPLING | BA_DBUS_PCM_UPDATE_CODEC);
-						bluealsa_dbus_pcm_update(&t_sco->sco.mic_pcm,
+						bluealsa_dbus_pcm_update(&t_sco->sco.pcm_mic,
 								BA_DBUS_PCM_UPDATE_SAMPLING | BA_DBUS_PCM_UPDATE_CODEC);
 					}
 				}
@@ -1312,9 +1342,9 @@ static void *rfcomm_thread(struct ba_rfcomm *r) {
 					/* If codec was selected during the SLC establishment,
 					 * notify BlueALSA D-Bus clients about the change. */
 					if (ba_transport_get_codec(t_sco) != HFP_CODEC_UNDEFINED) {
-						bluealsa_dbus_pcm_update(&t_sco->sco.spk_pcm,
+						bluealsa_dbus_pcm_update(&t_sco->sco.pcm_spk,
 								BA_DBUS_PCM_UPDATE_SAMPLING | BA_DBUS_PCM_UPDATE_CODEC);
-						bluealsa_dbus_pcm_update(&t_sco->sco.mic_pcm,
+						bluealsa_dbus_pcm_update(&t_sco->sco.pcm_mic,
 								BA_DBUS_PCM_UPDATE_SAMPLING | BA_DBUS_PCM_UPDATE_CODEC);
 					}
 				}
@@ -1358,7 +1388,7 @@ setup:
 					r->setup++;
 					break;
 				case HFP_SETUP_SELECT_CODEC:
-#if ENABLE_MSBC
+#if ENABLE_HFP_CODEC_SELECTION
 					if (r->idle) {
 						if (rfcomm_hfp_setup_codec_connection(r) == -1)
 							goto ioerror;
@@ -1378,7 +1408,7 @@ setup:
 					ba_transport_get_codec(t_sco) != HFP_CODEC_UNDEFINED)
 				r->setup = HFP_SETUP_COMPLETE;
 
-#if ENABLE_MSBC
+#if ENABLE_HFP_CODEC_SELECTION
 			/* Select HFP transport codec. Please note, that this setup
 			 * stage will be performed when the connection becomes idle. */
 			if (t_sco->profile & BA_TRANSPORT_PROFILE_HFP_AG &&
@@ -1427,7 +1457,7 @@ process:
 		if (pfds[0].revents & POLLIN) {
 			/* dispatch incoming event */
 			switch (rfcomm_recv_signal(r)) {
-#if ENABLE_MSBC
+#if ENABLE_HFP_CODEC_SELECTION
 			case BA_RFCOMM_SIGNAL_HFP_SET_CODEC_CVSD:
 				if (!config.hfp.codecs.cvsd || !(
 							r->ag_features & HFP_AG_FEAT_CODEC &&
@@ -1436,6 +1466,7 @@ process:
 				else if (rfcomm_hfp_set_codec(r, HFP_CODEC_CVSD) == -1)
 					goto ioerror;
 				break;
+# if ENABLE_MSBC
 			case BA_RFCOMM_SIGNAL_HFP_SET_CODEC_MSBC:
 				if (!config.hfp.codecs.msbc || !(
 							r->ag_features & HFP_AG_FEAT_CODEC &&
@@ -1446,6 +1477,19 @@ process:
 				else if (rfcomm_hfp_set_codec(r, HFP_CODEC_MSBC) == -1)
 					goto ioerror;
 				break;
+# endif
+# if ENABLE_LC3_SWB
+			case BA_RFCOMM_SIGNAL_HFP_SET_CODEC_LC3_SWB:
+				if (!config.hfp.codecs.lc3_swb || !(
+							r->ag_features & HFP_AG_FEAT_CODEC &&
+							r->ag_features & HFP_AG_FEAT_ESCO &&
+							r->hf_features & HFP_HF_FEAT_CODEC &&
+							r->hf_features & HFP_HF_FEAT_ESCO))
+					rfcomm_finalize_codec_selection(r);
+				else if (rfcomm_hfp_set_codec(r, HFP_CODEC_LC3_SWB) == -1)
+					goto ioerror;
+				break;
+# endif
 #endif
 			case BA_RFCOMM_SIGNAL_UPDATE_BATTERY:
 				if (rfcomm_notify_battery_level_change(r) == -1)
@@ -1577,7 +1621,7 @@ struct ba_rfcomm *ba_rfcomm_new(struct ba_transport *sco, int fd) {
 	r->thread = config.main_thread;
 	r->state = HFP_DISCONNECTED;
 	r->state_prev = HFP_DISCONNECTED;
-	r->codec = HFP_CODEC_UNDEFINED;
+	r->codec_id = HFP_CODEC_UNDEFINED;
 	r->sco = ba_transport_ref(sco);
 	r->link_lost_quirk = true;
 
@@ -1602,6 +1646,10 @@ struct ba_rfcomm *ba_rfcomm_new(struct ba_transport *sco, int fd) {
 #if ENABLE_MSBC
 		if (config.hfp.codecs.msbc && r->ag_features & HFP_AG_FEAT_ESCO)
 			r->ag_codecs.msbc = true;
+#endif
+#if ENABLE_LC3_SWB
+		if (config.hfp.codecs.lc3_swb && r->ag_features & HFP_AG_FEAT_ESCO)
+			r->ag_codecs.lc3_swb = true;
 #endif
 	}
 
@@ -1621,6 +1669,13 @@ struct ba_rfcomm *ba_rfcomm_new(struct ba_transport *sco, int fd) {
 			r->hf_codecs.msbc = true;
 		}
 #endif
+#if ENABLE_LC3_SWB
+		if (config.hfp.codecs.lc3_swb && r->hf_features & HFP_HF_FEAT_ESCO) {
+			const bool first = ptr == r->hf_bac_bcs_string;
+			ptr += sprintf(ptr, "%s%u", first ? "" : ",", HFP_CODEC_LC3_SWB);
+			r->hf_codecs.lc3_swb = true;
+		}
+#endif
 
 	}
 
@@ -1629,9 +1684,9 @@ struct ba_rfcomm *ba_rfcomm_new(struct ba_transport *sco, int fd) {
 
 	/* Initialize data used for volume gain synchronization. */
 	r->gain_mic = ba_transport_pcm_volume_level_to_range(
-			sco->sco.mic_pcm.volume[0].level, HFP_VOLUME_GAIN_MAX);
+			sco->sco.pcm_mic.volume[0].level, HFP_VOLUME_GAIN_MAX);
 	r->gain_spk = ba_transport_pcm_volume_level_to_range(
-			sco->sco.spk_pcm.volume[0].level, HFP_VOLUME_GAIN_MAX);
+			sco->sco.pcm_spk.volume[0].level, HFP_VOLUME_GAIN_MAX);
 
 	if (pipe(r->sig_fd) == -1)
 		goto fail;
