@@ -13,6 +13,9 @@
 #define BLUEALSA_BLUEALSAIFACE_H_
 
 #include <gio/gio.h>
+#include <glib.h>
+
+#include "dbus.h"
 
 #define BLUEALSA_SERVICE "org.bluealsa"
 
@@ -30,6 +33,7 @@
 #define BLUEALSA_TRANSPORT_TYPE_HSP         "HSP"
 #define BLUEALSA_TRANSPORT_TYPE_HSP_AG      BLUEALSA_TRANSPORT_TYPE_HSP "-AG"
 #define BLUEALSA_TRANSPORT_TYPE_HSP_HS      BLUEALSA_TRANSPORT_TYPE_HSP "-HS"
+#define BLUEALSA_TRANSPORT_TYPE_MIDI        "MIDI"
 
 #define BLUEALSA_PCM_CTRL_DRAIN  "Drain"
 #define BLUEALSA_PCM_CTRL_DROP   "Drop"
@@ -39,8 +43,28 @@
 #define BLUEALSA_PCM_MODE_SINK   "sink"
 #define BLUEALSA_PCM_MODE_SOURCE "source"
 
-extern const GDBusInterfaceInfo org_bluealsa_manager1_interface;
-extern const GDBusInterfaceInfo org_bluealsa_pcm1_interface;
-extern const GDBusInterfaceInfo org_bluealsa_rfcomm1_interface;
+typedef struct {
+	GDBusInterfaceSkeletonEx parent;
+} OrgBluealsaManager1Skeleton;
+
+OrgBluealsaManager1Skeleton *org_bluealsa_manager1_skeleton_new(
+		const GDBusInterfaceSkeletonVTable *vtable, void *userdata,
+		GDestroyNotify userdata_free_func);
+
+typedef struct {
+	GDBusInterfaceSkeletonEx parent;
+} OrgBluealsaPcm1Skeleton;
+
+OrgBluealsaPcm1Skeleton *org_bluealsa_pcm1_skeleton_new(
+		const GDBusInterfaceSkeletonVTable *vtable, void *userdata,
+		GDestroyNotify userdata_free_func);
+
+typedef struct {
+	GDBusInterfaceSkeletonEx parent;
+} OrgBluealsaRfcomm1Skeleton;
+
+OrgBluealsaRfcomm1Skeleton *org_bluealsa_rfcomm1_skeleton_new(
+		const GDBusInterfaceSkeletonVTable *vtable, void *userdata,
+		GDestroyNotify userdata_free_func);
 
 #endif
