@@ -1,6 +1,6 @@
 /*
  * BlueALSA - codec-msbc.c
- * Copyright (c) 2016-2022 Arkadiusz Bokowy
+ * Copyright (c) 2016-2024 Arkadiusz Bokowy
  * Copyright (c) 2017 Juha Kuikka
  *
  * This file is a part of bluez-alsa.
@@ -10,17 +10,19 @@
  */
 
 #include "codec-msbc.h"
-/* IWYU pragma: no_include "config.h" */
 
-#include <endian.h>
+#if HAVE_CONFIG_H
+# include <config.h>
+#endif
+
 #include <errno.h>
 #include <stdbool.h>
 #include <stdint.h>
-#include <string.h>
 
 #include <spandsp.h>
 
 #include "codec-sbc.h"
+#include "h2.h"
 #include "shared/log.h"
 
 /**
@@ -32,67 +34,30 @@
  * correct number of PCM samples - matching sampling frequency. */
 #define MSBC_DECODE_ERROR_PLC 1
 
-/* Code protected 2-bit sequence numbers (SN0 and SN1) used
- * by the msbc_encode() function. */
-static const uint8_t sn[][2] = {
-	{ 0, 0 }, { 3, 0 }, { 0, 3 }, { 3, 3 }
-};
-
 /**
- * Find H2 synchronization header within eSCO transparent data.
+ * Initialize mSBC codec structure.
  *
- * @param data Memory area with the eSCO transparent data.
- * @param len Address from where the length of the eSCO transparent data
- *   is read. Upon exit, the remaining length of the eSCO data will be
- *   stored in this variable (received length minus scanned length).
- * @return On success this function returns address of the first occurrence
- *   of the H2 synchronization header. Otherwise, it returns NULL. */
-static void *msbc_find_h2_header(const void *data, size_t *len) {
-
-	const uint8_t *_data = data;
-	size_t _len = *len;
-	void *ptr = NULL;
-
-	while (_len >= sizeof(esco_h2_header_t)) {
-
-		esco_h2_header_t h2;
-		memcpy(&h2, _data, sizeof(h2));
-		h2 = le16toh(h2);
-
-		if (ESCO_H2_GET_SYNCWORD(h2) == ESCO_H2_SYNCWORD &&
-				(ESCO_H2_GET_SN0(h2) >> 1) == (ESCO_H2_GET_SN0(h2) & 1) &&
-				(ESCO_H2_GET_SN1(h2) >> 1) == (ESCO_H2_GET_SN1(h2) & 1)) {
-			ptr = (void *)_data;
-			goto final;
-		}
-
-		_data += 1;
-		_len--;
-	}
-
-final:
-	*len = _len;
-	return ptr;
-}
-
+ * This function is idempotent, so it can be called multiple times in order
+ * to reinitialize the codec structure.
+ *
+ * @param msbc Codec structure which shall be initialized.
+ * @return This function returns 0 on success or a negative error value
+ *   in case of initialization failure. */
 int msbc_init(struct esco_msbc *msbc) {
 
-	int err;
-
 	if (!msbc->initialized) {
 		debug("Initializing mSBC codec");
 		if ((errno = -sbc_init_msbc(&msbc->sbc, 0)) != 0)
-			goto fail;
-		if (ffb_init_uint8_t(&msbc->data, sizeof(esco_msbc_frame_t) * 3) == -1)
-			goto fail;
-		/* Allocate buffer for 1 decoded frame, optional 3 PLC frames and
-		 * some extra frames to account for async PCM samples reading. */
-		if (ffb_init_int16_t(&msbc->pcm, MSBC_CODESAMPLES * 6) == -1)
-			goto fail;
+			return -errno;
+	}
+	else {
+		debug("Re-initializing mSBC codec");
+		if ((errno = -sbc_reinit_msbc(&msbc->sbc, 0)) != 0)
+			return -errno;
 	}
 
-	if ((errno = -sbc_reinit_msbc(&msbc->sbc, 0)) != 0)
-		return -1;
+	/* ensure libsbc uses little-endian PCM on all architectures */
+	msbc->sbc.endian = SBC_LE;
 
 #if DEBUG
 	size_t len;
@@ -108,8 +73,8 @@ int msbc_init(struct esco_msbc *msbc) {
 	}
 #endif
 
-	ffb_rewind(&msbc->data);
-	ffb_rewind(&msbc->pcm);
+	ffb_init_from_array(&msbc->data, msbc->buffer_data);
+	ffb_init_from_array(&msbc->pcm, msbc->buffer_pcm);
 
 	msbc->seq_initialized = false;
 	msbc->seq_number = 0;
@@ -124,10 +89,8 @@ int msbc_init(struct esco_msbc *msbc) {
 	return 0;
 
 fail:
-	err = errno;
-	msbc_finish(msbc);
-	errno = err;
-	return -1;
+	sbc_finish(&msbc->sbc);
+	return -errno;
 }
 
 void msbc_finish(struct esco_msbc *msbc) {
@@ -137,9 +100,6 @@ void msbc_finish(struct esco_msbc *msbc) {
 
 	sbc_finish(&msbc->sbc);
 
-	ffb_free(&msbc->data);
-	ffb_free(&msbc->pcm);
-
 	plc_free(msbc->plc);
 	msbc->plc = NULL;
 
@@ -158,7 +118,7 @@ ssize_t msbc_decode(struct esco_msbc *msbc) {
 	ssize_t rv = 0;
 
 	const size_t tmp = input_len;
-	const esco_msbc_frame_t *frame = msbc_find_h2_header(input, &input_len);
+	const h2_msbc_frame_t *frame = h2_header_find(input, &input_len);
 	input += tmp - input_len;
 
 	/* Skip decoding if there is not enough input data or the output
@@ -168,24 +128,21 @@ ssize_t msbc_decode(struct esco_msbc *msbc) {
 			output_len < MSBC_CODESIZE * (1 + 3))
 		goto final;
 
-	esco_h2_header_t h2;
-	memcpy(&h2, frame, sizeof(h2));
-	h2 = le16toh(h2);
+	const uint8_t h2_seq = h2_header_unpack(frame->header);
 
-	uint8_t _seq = (ESCO_H2_GET_SN1(h2) & 2) | (ESCO_H2_GET_SN0(h2) & 1);
 	if (!msbc->seq_initialized) {
 		msbc->seq_initialized = true;
-		msbc->seq_number = _seq;
+		msbc->seq_number = h2_seq;
 	}
-	else if (_seq != ++msbc->seq_number) {
+	else if (h2_seq != ++msbc->seq_number) {
 
 		/* In case of missing mSBC frames (we can detect up to 3 consecutive
 		 * missing frames) use PLC for PCM samples reconstruction. */
 
-		uint8_t missing = (_seq + ESCO_H2_SN_MAX - msbc->seq_number) % ESCO_H2_SN_MAX;
-		warn("Missing mSBC packets (%u != %u): %u", _seq, msbc->seq_number, missing);
+		uint8_t missing = (h2_seq + 4 - msbc->seq_number) % 4;
+		warn("Missing mSBC packets (%u != %u): %u", h2_seq, msbc->seq_number, missing);
 
-		msbc->seq_number = _seq;
+		msbc->seq_number = h2_seq;
 
 		plc_fillin(msbc->plc, msbc->pcm.tail, missing * MSBC_CODESAMPLES);
 		ffb_seek(&msbc->pcm, missing * MSBC_CODESAMPLES);
@@ -197,17 +154,16 @@ ssize_t msbc_decode(struct esco_msbc *msbc) {
 	if ((len = sbc_decode(&msbc->sbc, frame->payload, sizeof(frame->payload),
 					msbc->pcm.tail, output_len, NULL)) < 0) {
 
+		warn("Couldn't decode mSBC frame: %s", sbc_strerror(len));
+
 		/* Move forward one byte to avoid getting stuck in
 		 * decoding the same mSBC packet all over again. */
 		input += 1;
 
 #if MSBC_DECODE_ERROR_PLC
-
-		warn("Couldn't decode mSBC frame: %s", sbc_strerror(len));
 		plc_fillin(msbc->plc, msbc->pcm.tail, MSBC_CODESAMPLES);
 		ffb_seek(&msbc->pcm, MSBC_CODESAMPLES);
 		rv += MSBC_CODESAMPLES;
-
 #else
 		rv = len;
 #endif
@@ -237,7 +193,7 @@ ssize_t msbc_encode(struct esco_msbc *msbc) {
 
 	const int16_t *input = msbc->pcm.data;
 	const size_t input_len = ffb_blen_out(&msbc->pcm);
-	esco_msbc_frame_t *frame = (esco_msbc_frame_t *)msbc->data.tail;
+	h2_msbc_frame_t *frame = msbc->data.tail;
 	size_t output_len = ffb_blen_in(&msbc->data);
 
 	/* Skip encoding if there is not enough PCM samples or the output
@@ -251,15 +207,23 @@ ssize_t msbc_encode(struct esco_msbc *msbc) {
 					frame->payload, sizeof(frame->payload), NULL)) < 0)
 		return len;
 
-	const uint8_t n = msbc->seq_number++;
-	frame->header = htole16(ESCO_H2_PACK(sn[n][0], sn[n][1]));
+	frame->header = h2_header_pack(++msbc->seq_number);
 	frame->padding = 0;
 
 	ffb_seek(&msbc->data, sizeof(*frame));
 	msbc->frames++;
 
 	/* Reshuffle remaining PCM data to the beginning of the buffer. */
-	ffb_shift(&msbc->pcm, input + MSBC_CODESAMPLES - (int16_t *)msbc->pcm.data);
+	ffb_shift(&msbc->pcm, MSBC_CODESAMPLES);
 
 	return sizeof(*frame);
 }
+
+/**
+ * Get string representation of the mSBC encode/decode error.
+ *
+ * @param err The encode/decode error code.
+ * @return Human-readable string. */
+const char *msbc_strerror(int err) {
+	return sbc_strerror(err);
+}
