@@ -1,6 +1,6 @@
 /*
  * BlueALSA - codec-msbc.h
- * Copyright (c) 2016-2022 Arkadiusz Bokowy
+ * Copyright (c) 2016-2024 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -24,6 +24,7 @@
 #include <sbc/sbc.h>
 #include <spandsp.h>
 
+#include "h2.h"
 #include "shared/ffb.h"
 
 /* HFP uses SBC encoding with precisely defined parameters. Hence, the size
@@ -32,22 +33,11 @@
 #define MSBC_CODESAMPLES (MSBC_CODESIZE / sizeof(int16_t))
 #define MSBC_FRAMELEN    57
 
-#define ESCO_H2_SYNCWORD 0x801
-#define ESCO_H2_SN_MAX   0x4
-#define ESCO_H2_GET_SYNCWORD(h2) ((h2) & 0xFFF)
-#define ESCO_H2_GET_SN0(h2)      (((h2) >> 12) & 0x3)
-#define ESCO_H2_GET_SN1(h2)      (((h2) >> 14) & 0x3)
-/* Pack two repetition code protected 2-bit sequence numbers (both bits
- * duplicated) into the 16-bit eSCO H2 header. Note, that after packing,
- * the H2 header value has to be converted to little-endian. */
-#define ESCO_H2_PACK(sn0, sn1) (ESCO_H2_SYNCWORD | (sn0) << 12 | (sn1) << 14)
-
-typedef uint16_t esco_h2_header_t;
-typedef struct esco_msbc_frame {
-	esco_h2_header_t header;
+typedef struct h2_msbc_frame {
+	h2_header_t header;
 	uint8_t payload[MSBC_FRAMELEN];
 	uint8_t padding;
-} __attribute__ ((packed)) esco_msbc_frame_t;
+} __attribute__ ((packed)) h2_msbc_frame_t;
 
 struct esco_msbc {
 
@@ -59,7 +49,7 @@ struct esco_msbc {
 	/* buffer for PCM samples */
 	ffb_t pcm;
 
-	uint8_t seq_initialized : 1;
+	bool seq_initialized;
 	uint8_t seq_number : 2;
 	/* number of processed frames */
 	size_t frames;
@@ -71,6 +61,13 @@ struct esco_msbc {
 	 * used for reinitialization - it makes msbc_init() idempotent. */
 	bool initialized;
 
+	/* Allocated buffer for 3 mSBC frames to have some extra space in case of
+	 * PCM samples asynchronous reading beeing slower than incoming frames. */
+	uint8_t buffer_data[sizeof(h2_msbc_frame_t) * 3];
+	/* Allocate buffer for 1 decoded frame, optional 3 PLC frames and
+	 * some extra frames to account for async PCM samples reading. */
+	int16_t buffer_pcm[MSBC_CODESAMPLES * 6];
+
 };
 
 int msbc_init(struct esco_msbc *msbc);
@@ -79,4 +76,6 @@ void msbc_finish(struct esco_msbc *msbc);
 ssize_t msbc_decode(struct esco_msbc *msbc);
 ssize_t msbc_encode(struct esco_msbc *msbc);
 
+const char *msbc_strerror(int err);
+
 #endif
