@@ -1,6 +1,6 @@
 /*
  * BlueALSA - io.c
- * Copyright (c) 2016-2023 Arkadiusz Bokowy
+ * Copyright (c) 2016-2024 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -22,18 +22,19 @@
 #include <glib.h>
 
 #include "audio.h"
-#include "bluealsa-config.h"
+#include "ba-config.h"
 #include "shared/defs.h"
+#include "shared/ffb.h"
 #include "shared/log.h"
 
 /**
  * Read data from the BT transport (SCO or SEQPACKET) socket. */
 ssize_t io_bt_read(
-		struct ba_transport_thread *th,
+		struct ba_transport_pcm *pcm,
 		void *buffer,
 		size_t count) {
 
-	const int fd = th->bt_fd;
+	const int fd = pcm->fd_bt;
 	ssize_t ret;
 
 retry:
@@ -53,7 +54,7 @@ retry:
 	}
 
 	if (ret == 0)
-		ba_transport_thread_bt_release(th);
+		ba_transport_pcm_bt_release(pcm);
 
 	return ret;
 }
@@ -64,11 +65,11 @@ retry:
  * Note:
  * This function may temporally re-enable thread cancellation! */
 ssize_t io_bt_write(
-		struct ba_transport_thread *th,
+		struct ba_transport_pcm *pcm,
 		const void *buffer,
 		size_t count) {
 
-	const int fd = th->bt_fd;
+	const int fd = pcm->fd_bt;
 	ssize_t ret;
 
 retry:
@@ -96,7 +97,7 @@ retry:
 		}
 
 	if (ret == 0)
-		ba_transport_thread_bt_release(th);
+		ba_transport_pcm_bt_release(pcm);
 
 	return ret;
 }
@@ -112,42 +113,42 @@ void io_pcm_scale(
 
 	const unsigned int channels = pcm->channels;
 	const bool pcm_soft_volume = pcm->soft_volume;
-	const double pcm_volume_ch_scales[2] = {
+	double pcm_volume_ch_scales[] = {
 		pcm->volume[0].scale,
 		pcm->volume[1].scale,
 	};
 
 	pthread_mutex_unlock(&pcm->mutex);
 
-	if (!pcm_soft_volume) {
+	g_assert_cmpint(channels, <=, ARRAYSIZE(pcm_volume_ch_scales));
+
+	if (!pcm_soft_volume)
 		/* In case of hardware volume control we will perform mute operation,
 		 * because hardware muting is an equivalent of gain=0 which with some
 		 * headsets does not entirely silence audio. */
-		switch (pcm->format) {
-		case BA_TRANSPORT_PCM_FORMAT_S16_2LE:
-			audio_silence_s16_2le(buffer, samples / channels, channels,
-					pcm_volume_ch_scales[0] == 0, pcm_volume_ch_scales[1] == 0);
+		for (size_t i = 0; i < channels; i++)
+			if (pcm_volume_ch_scales[i] != 0.0)
+				pcm_volume_ch_scales[i] = 1.0;
+
+	bool identity = true;
+	/* Check whether volume scaling is required. */
+	for (size_t i = 0; i < channels; i++)
+		if (pcm_volume_ch_scales[i] != 1.0) {
+			identity = false;
 			break;
-		case BA_TRANSPORT_PCM_FORMAT_S24_4LE:
-		case BA_TRANSPORT_PCM_FORMAT_S32_4LE:
-			audio_silence_s32_4le(buffer, samples / channels, channels,
-					pcm_volume_ch_scales[0] == 0, pcm_volume_ch_scales[1] == 0);
-			break;
-		default:
-			g_assert_not_reached();
 		}
+
+	if (identity)
+		/* Nothing to do - volume is set to 100%. */
 		return;
-	}
 
 	switch (pcm->format) {
 	case BA_TRANSPORT_PCM_FORMAT_S16_2LE:
-		audio_scale_s16_2le(buffer, samples / channels, channels,
-				pcm_volume_ch_scales[0], pcm_volume_ch_scales[1]);
+		audio_scale_s16_2le(buffer, pcm_volume_ch_scales, channels, samples / channels);
 		break;
 	case BA_TRANSPORT_PCM_FORMAT_S24_4LE:
 	case BA_TRANSPORT_PCM_FORMAT_S32_4LE:
-		audio_scale_s32_4le(buffer, samples / channels, channels,
-				pcm_volume_ch_scales[0], pcm_volume_ch_scales[1]);
+		audio_scale_s32_4le(buffer, pcm_volume_ch_scales, channels, samples / channels);
 		break;
 	default:
 		g_assert_not_reached();
@@ -238,6 +239,7 @@ ssize_t io_pcm_write(
 				 * dropped in the bluetooth controller if we block here.
 				 * It is better that we discard frames here so that the
 				 * decoder is not interrupted. */
+				warn("Dropping PCM frames: %s", "PCM overrun");
 				ret = len;
 				break;
 			case EPIPE:
@@ -264,13 +266,6 @@ final:
 	return ret;
 }
 
-static enum ba_transport_thread_signal io_poll_signal_filter_none(
-		enum ba_transport_thread_signal signal,
-		void *userdata) {
-	(void)userdata;
-	return signal;
-}
-
 /**
  * Poll and read data from the BT transport socket.
  *
@@ -278,13 +273,12 @@ static enum ba_transport_thread_signal io_poll_signal_filter_none(
  * This function temporally re-enables thread cancellation! */
 ssize_t io_poll_and_read_bt(
 		struct io_poll *io,
-		struct ba_transport_thread *th,
-		void *buffer,
-		size_t count) {
+		struct ba_transport_pcm *pcm,
+		ffb_t *buffer) {
 
-	struct pollfd fds[2] = {
-		{ th->pipe[0], POLLIN, 0 },
-		{ th->bt_fd, POLLIN, 0 }};
+	struct pollfd fds[] = {
+		{ pcm->pipe[0], POLLIN, 0 },
+		{ pcm->fd_bt, POLLIN, 0 }};
 
 repoll:
 
@@ -298,19 +292,16 @@ repoll:
 		return -1;
 	}
 
-	if (fds[0].revents & POLLIN) {
-		/* dispatch incoming event */
-		io_poll_signal_filter *filter = io->signal.filter != NULL ?
-			io->signal.filter : io_poll_signal_filter_none;
-		enum ba_transport_thread_signal signal;
-		ba_transport_thread_signal_recv(th, &signal);
-		switch (filter(signal, io->signal.userdata)) {
+	if (fds[0].revents & POLLIN)
+		switch (ba_transport_pcm_signal_recv(pcm)) {
 		default:
 			goto repoll;
 		}
-	}
 
-	return io_bt_read(th, buffer, count);
+	ssize_t len;
+	if ((len = io_bt_read(pcm, buffer->tail, ffb_blen_in(buffer))) > 0)
+		ffb_seek(buffer, len);
+	return len;
 }
 
 /**
@@ -321,19 +312,17 @@ repoll:
 ssize_t io_poll_and_read_pcm(
 		struct io_poll *io,
 		struct ba_transport_pcm *pcm,
-		void *buffer,
-		size_t samples) {
+		ffb_t *buffer) {
 
-	struct ba_transport_thread *th = pcm->th;
-	struct pollfd fds[2] = {
-		{ th->pipe[0], POLLIN, 0 },
+	struct pollfd fds[] = {
+		{ pcm->pipe[0], POLLIN, 0 },
 		{ -1, POLLIN, 0 }};
 
 repoll:
 
 	pthread_mutex_lock(&pcm->mutex);
-	/* Add PCM socket to the poll if it is active. */
-	fds[1].fd = pcm->active ? pcm->fd : -1;
+	/* Add PCM socket to the poll if it is not paused. */
+	fds[1].fd = pcm->paused ? -1 : pcm->fd;
 	pthread_mutex_unlock(&pcm->mutex);
 
 	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
@@ -355,25 +344,20 @@ repoll:
 		return -1;
 	}
 
-	if (fds[0].revents & POLLIN) {
-		/* dispatch incoming event */
-		io_poll_signal_filter *filter = io->signal.filter != NULL ?
-			io->signal.filter : io_poll_signal_filter_none;
-		enum ba_transport_thread_signal signal;
-		ba_transport_thread_signal_recv(th, &signal);
-		switch (filter(signal, io->signal.userdata)) {
-		case BA_TRANSPORT_THREAD_SIGNAL_PCM_OPEN:
-		case BA_TRANSPORT_THREAD_SIGNAL_PCM_RESUME:
+	if (fds[0].revents & POLLIN)
+		switch (ba_transport_pcm_signal_recv(pcm)) {
+		case BA_TRANSPORT_PCM_SIGNAL_OPEN:
+		case BA_TRANSPORT_PCM_SIGNAL_RESUME:
 			io->asrs.frames = 0;
 			io->timeout = -1;
 			goto repoll;
-		case BA_TRANSPORT_THREAD_SIGNAL_PCM_CLOSE:
+		case BA_TRANSPORT_PCM_SIGNAL_CLOSE:
 			/* reuse PCM read disconnection logic */
 			break;
-		case BA_TRANSPORT_THREAD_SIGNAL_PCM_SYNC:
+		case BA_TRANSPORT_PCM_SIGNAL_SYNC:
 			io->timeout = 100;
 			goto repoll;
-		case BA_TRANSPORT_THREAD_SIGNAL_PCM_DROP:
+		case BA_TRANSPORT_PCM_SIGNAL_DROP:
 			/* Notify caller that the PCM FIFO has been dropped. This will give
 			 * the caller a chance to reinitialize its internal state. */
 			errno = ESTALE;
@@ -381,21 +365,20 @@ repoll:
 		default:
 			goto repoll;
 		}
-	}
 
 	if (fds[1].revents == 0)
 		return 0;
 
-	ssize_t samples_read;
-	if ((samples_read = io_pcm_read(pcm, buffer, samples)) == -1) {
+	ssize_t samples;
+	if ((samples = io_pcm_read(pcm, buffer->tail, ffb_len_in(buffer))) == -1) {
 		if (errno == EAGAIN)
 			goto repoll;
 		if (errno != EBADF)
 			return -1;
-		samples_read = 0;
+		samples = 0;
 	}
 
-	if (samples_read == 0)
+	if (samples == 0)
 		return 0;
 
 	/* When the thread is created, there might be no data in the FIFO. In fact
@@ -405,5 +388,6 @@ repoll:
 	if (io->asrs.frames == 0)
 		asrsync_init(&io->asrs, pcm->sampling);
 
-	return samples_read;
+	ffb_seek(buffer, samples);
+	return samples;
 }
