@@ -1,6 +1,6 @@
 /*
  * BlueALSA - io.h
- * Copyright (c) 2016-2023 Arkadiusz Bokowy
+ * Copyright (c) 2016-2024 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -18,26 +18,16 @@
 
 #include <sys/types.h>
 
-#include "ba-transport.h"
+#include "ba-transport-pcm.h"
+#include "shared/ffb.h"
 #include "shared/rt.h"
 
-/**
- * Callback function for thread signal filtering. */
-typedef enum ba_transport_thread_signal io_poll_signal_filter(
-		enum ba_transport_thread_signal signal,
-		void *userdata);
-
 /**
  * Data associated with IO polling.
  *
  * Note:
  * The timeout field shall be initialized to -1. */
 struct io_poll {
-	struct {
-		/* signal filtering callback */
-		io_poll_signal_filter *filter;
-		void *userdata;
-	} signal;
 	/* transfer bit rate synchronization */
 	struct asrsync asrs;
 	/* keep-alive and sync timeout */
@@ -45,12 +35,12 @@ struct io_poll {
 };
 
 ssize_t io_bt_read(
-		struct ba_transport_thread *th,
+		struct ba_transport_pcm *pcm,
 		void *buffer,
 		size_t count);
 
 ssize_t io_bt_write(
-		struct ba_transport_thread *th,
+		struct ba_transport_pcm *pcm,
 		const void *buffer,
 		size_t count);
 
@@ -74,14 +64,12 @@ ssize_t io_pcm_write(
 
 ssize_t io_poll_and_read_bt(
 		struct io_poll *io,
-		struct ba_transport_thread *th,
-		void *buffer,
-		size_t count);
+		struct ba_transport_pcm *pcm,
+		ffb_t *buffer);
 
 ssize_t io_poll_and_read_pcm(
 		struct io_poll *io,
 		struct ba_transport_pcm *pcm,
-		void *buffer,
-		size_t samples);
+		ffb_t *buffer);
 
 #endif
