@@ -1,7 +1,7 @@
 /*
  * BlueALSA - ofono.h
  * Copyright (c) 2016-2022 Arkadiusz Bokowy
- *               2018 Thierry Bultel
+ * Copyright (c) 2018 Thierry Bultel
  *
  * This file is a part of bluez-alsa.
  *
@@ -13,9 +13,17 @@
 #ifndef BLUEALSA_OFONO_H_
 #define BLUEALSA_OFONO_H_
 
+#if HAVE_CONFIG_H
+# include <config.h>
+#endif
+
 #include <stdbool.h>
 
+#include "ba-transport.h"
+
 int ofono_init(void);
 bool ofono_detect_service(void);
 
+int ofono_call_volume_update(struct ba_transport *t);
+
 #endif
