@@ -1,6 +1,6 @@
 /*
  * BlueALSA - bluetooth.h
- * Copyright (c) 2016-2023 Arkadiusz Bokowy
+ * Copyright (c) 2016-2024 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -24,19 +24,25 @@
 #define BT_COMPID_REALTEK            0x005D
 #define BT_COMPID_SAMSUNG_ELEC       0x0075
 #define BT_COMPID_QUALCOMM_TECH      0x00D7
+#define BT_COMPID_GOOGLE             0x00E0
 #define BT_COMPID_SONY               0x012D
 #define BT_COMPID_CYPRESS            0x0131
 #define BT_COMPID_SAVITECH           0x053A
+#define BT_COMPID_LINUX_FOUNDATION   0x05F1
 #define BT_COMPID_FRAUNHOFER_IIS     0x08A9
 
 /**
  * Bluetooth UUIDs associated with audio profiles. */
 
-#define BT_UUID_A2DP_SOURCE "0000110A-0000-1000-8000-00805F9B34FB"
-#define BT_UUID_A2DP_SINK   "0000110B-0000-1000-8000-00805F9B34FB"
-#define BT_UUID_HSP_HS      "00001108-0000-1000-8000-00805F9B34FB"
-#define BT_UUID_HSP_AG      "00001112-0000-1000-8000-00805F9B34FB"
-#define BT_UUID_HFP_HF      "0000111E-0000-1000-8000-00805F9B34FB"
-#define BT_UUID_HFP_AG      "0000111F-0000-1000-8000-00805F9B34FB"
+#define BT_UUID_A2DP_SOURCE "0000110a-0000-1000-8000-00805f9b34fb"
+#define BT_UUID_A2DP_SINK   "0000110b-0000-1000-8000-00805f9b34fb"
+#define BT_UUID_HSP_HS      "00001108-0000-1000-8000-00805f9b34fb"
+#define BT_UUID_HSP_AG      "00001112-0000-1000-8000-00805f9b34fb"
+#define BT_UUID_HFP_HF      "0000111e-0000-1000-8000-00805f9b34fb"
+#define BT_UUID_HFP_AG      "0000111f-0000-1000-8000-00805f9b34fb"
+
+#define BT_UUID_MIDI        "03b80e5a-ede8-4b33-a751-6ce34ec4c700"
+#define BT_UUID_MIDI_CHAR   "7772e5db-3868-4112-a1a9-f2669d106bf3"
+#define BT_UUID_MIDI_DESC   "00002901-0000-1000-8000-00805f9b34fb"
 
 #endif
