@@ -111,7 +111,7 @@ static void vlog(int priority, const char *format, va_list ap) {
 			fprintf(stderr, "%s: ", _ident);
 
 #if DEBUG_TIME
-		fprintf(stderr, "%lu.%.9lu: ", (long int)ts.tv_sec, ts.tv_nsec);
+		fprintf(stderr, "%lu.%.6lu: ", (long int)ts.tv_sec, ts.tv_nsec / 1000);
 #endif
 
 #if DEBUG && HAVE_GETTID
@@ -172,8 +172,7 @@ void callstackdump_(const char *label) {
 	size_t n = backtrace(frames, ARRAYSIZE(frames));
 	char **symbols = backtrace_symbols(frames, n);
 
-	size_t i;
-	for (i = 1; i < n; i++)
+	for (size_t i = 1; i < n; i++)
 		ptr += snprintf(ptr, sizeof(buffer) + buffer - ptr, "%s%s",
 				symbols[i], (i + 1 < n) ? " < " : "");
 
@@ -192,17 +191,14 @@ void callstackdump_(const char *label) {
  *
  * @param label Label printed before the memory block output.
  * @param data Address of the memory block.
- * @param len Number of bytes which should be printed.
- * @param compact Dump memory block without spaces. */
-void hexdump_(const char *label, const void *data, size_t len, bool compact) {
+ * @param len Number of bytes which should be printed. */
+void hexdump_(const char *label, const void *data, size_t len) {
 
-	const char *sep = "";
-	const char *spacing = compact ? "" : " ";
 	char *buf = malloc(len * 3 + 1);
 	char *p = buf;
 
-	for (size_t i = 0; i < len; sep = ++i % 4 ? "" : spacing) {
-		p += sprintf(p, "%s%02x", sep, *(unsigned char *)data & 0xFF);
+	for (size_t i = 0; i < len; i++) {
+		p += sprintf(p, "%02x", *(unsigned char *)data & 0xFF);
 		data = ((unsigned char *)data) + 1;
 	}
 
