@@ -1,6 +1,6 @@
 /*
  * BlueALSA - utils.h
- * Copyright (c) 2016-2020 Arkadiusz Bokowy
+ * Copyright (c) 2016-2024 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -8,6 +8,7 @@
  *
  */
 
+#pragma once
 #ifndef BLUEALSA_UTILS_H_
 #define BLUEALSA_UTILS_H_
 
@@ -22,21 +23,19 @@
 
 #include <glib.h>
 
-#include "ba-transport.h"
-
 int g_dbus_bluez_object_path_to_hci_dev_id(const char *path);
 bdaddr_t *g_dbus_bluez_object_path_to_bdaddr(const char *path, bdaddr_t *addr);
-const char *g_dbus_transport_type_to_bluez_object_path(struct ba_transport_type type);
 
 char *g_variant_sanitize_object_path(char *path);
 bool g_variant_validate_value(GVariant *value, const GVariantType *type,
 		const char *name);
 
+GSource *g_io_create_watch_full(GIOChannel *channel, int priority,
+		GIOCondition cond, GIOFunc func, void *userdata, GDestroyNotify notify);
+
 unsigned int g_bdaddr_hash(const void *v);
 gboolean g_bdaddr_equal(const void *v1, const void *v2);
 
-const char *ba_transport_type_to_string(struct ba_transport_type type);
-
 #if ENABLE_MP3LAME
 int a2dp_mpeg1_mp3_get_max_bitrate(uint16_t mask);
 const char *lame_encode_strerror(int err);
@@ -50,7 +49,7 @@ const char *aacenc_strerror(AACENC_ERROR err);
 #endif
 
 #if ENABLE_LC3PLUS
-# include <lc3.h>
+# include <lc3plus.h>
 const char *lc3plus_strerror(LC3PLUS_Error err);
 #endif
 
