@@ -1,6 +1,6 @@
 /*
  * test-alsa-ctl.c
- * Copyright (c) 2016-2023 Arkadiusz Bokowy
+ * Copyright (c) 2016-2024 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -22,6 +22,8 @@
 #include <check.h>
 #include <alsa/asoundlib.h>
 
+#include "shared/log.h"
+
 #include "inc/check.inc"
 #include "inc/mock.inc"
 #include "inc/preload.inc"
@@ -49,6 +51,25 @@ static int test_pcm_close(struct spawn_process *sp_ba_mock, snd_ctl_t *ctl) {
 	return rv;
 }
 
+#if DEBUG
+static const char *test_ctl_event_elem_get_mask_name(snd_ctl_event_t *event) {
+	switch (snd_ctl_event_elem_get_mask(event)) {
+	case SND_CTL_EVENT_MASK_ADD:
+		return "ADD";
+	case SND_CTL_EVENT_MASK_REMOVE:
+		return "REMOVE";
+	case SND_CTL_EVENT_MASK_VALUE:
+		return "VALUE";
+	case SND_CTL_EVENT_MASK_INFO:
+		return "INFO";
+	case SND_CTL_EVENT_MASK_TLV:
+		return "TLV";
+	default:
+		return "UNKNOWN";
+	}
+}
+#endif
+
 CK_START_TEST(test_controls) {
 
 	struct spawn_process sp_ba_mock;
@@ -96,7 +117,7 @@ CK_START_TEST(test_controls_battery) {
 				"--profile=hsp-ag",
 				NULL), -1);
 
-	ck_assert_int_eq(snd_ctl_open(&ctl, "bluealsa:BAT=yes", 0), 0);
+	ck_assert_int_eq(snd_ctl_open(&ctl, "bluealsa:EXT=battery", 0), 0);
 
 	snd_ctl_elem_list_t *elems;
 	snd_ctl_elem_list_alloca(&elems);
@@ -142,27 +163,30 @@ CK_START_TEST(test_controls_extended) {
 	snd_ctl_elem_list_alloca(&elems);
 
 	ck_assert_int_eq(snd_ctl_elem_list(ctl, elems), 0);
-	ck_assert_int_eq(snd_ctl_elem_list_get_count(elems), 15);
-	ck_assert_int_eq(snd_ctl_elem_list_alloc_space(elems, 15), 0);
+	ck_assert_int_eq(snd_ctl_elem_list_get_count(elems), 20);
+	ck_assert_int_eq(snd_ctl_elem_list_alloc_space(elems, 20), 0);
 	ck_assert_int_eq(snd_ctl_elem_list(ctl, elems), 0);
 
 	/* codec control element shall be after playback/capture elements */
 	ck_assert_str_eq(snd_ctl_elem_list_get_name(elems, 3), "12:34:56:78:9A:BC A2DP Codec Enum");
-	ck_assert_str_eq(snd_ctl_elem_list_get_name(elems, 10), "12:34:56:78:9A:BC SCO Codec Enum");
-	ck_assert_str_eq(snd_ctl_elem_list_get_name(elems, 14), "23:45:67:89:AB:CD A2DP Codec Enum");
+	ck_assert_str_eq(snd_ctl_elem_list_get_name(elems, 11), "12:34:56:78:9A:BC SCO Codec Enum");
+	ck_assert_str_eq(snd_ctl_elem_list_get_name(elems, 18), "23:45:67:89:AB:CD A2DP Codec Enum");
 
-	bool has_msbc = false;
+	int sco_codec_enum_items = 1;
 #if ENABLE_MSBC
-	has_msbc = true;
+	sco_codec_enum_items += 1;
+#endif
+#if ENABLE_LC3_SWB
+	sco_codec_enum_items += 1;
 #endif
 
 	snd_ctl_elem_info_t *info;
 	snd_ctl_elem_info_alloca(&info);
 
 	/* 12:34:56:78:9A:BC SCO Codec Enum */
-	snd_ctl_elem_info_set_numid(info, snd_ctl_elem_list_get_numid(elems, 10));
+	snd_ctl_elem_info_set_numid(info, snd_ctl_elem_list_get_numid(elems, 11));
 	ck_assert_int_eq(snd_ctl_elem_info(ctl, info), 0);
-	ck_assert_int_eq(snd_ctl_elem_info_get_items(info), has_msbc ? 2 : 1);
+	ck_assert_int_eq(snd_ctl_elem_info_get_items(info), sco_codec_enum_items);
 	snd_ctl_elem_info_set_item(info, 0);
 	ck_assert_int_eq(snd_ctl_elem_info(ctl, info), 0);
 	ck_assert_str_eq(snd_ctl_elem_info_get_item_name(info), "CVSD");
@@ -186,10 +210,11 @@ CK_START_TEST(test_controls_extended) {
 	ck_assert_int_eq(snd_ctl_elem_write(ctl, elem), 0);
 
 	/* 12:34:56:78:9A:BC SCO Codec Enum */
-	snd_ctl_elem_value_set_numid(elem, snd_ctl_elem_list_get_numid(elems, 10));
+	snd_ctl_elem_value_set_numid(elem, snd_ctl_elem_list_get_numid(elems, 11));
 	/* get currently selected SCO codec */
 	ck_assert_int_eq(snd_ctl_elem_read(ctl, elem), 0);
-	ck_assert_int_eq(snd_ctl_elem_value_get_enumerated(elem, 0), has_msbc ? 1 : 0);
+	ck_assert_int_eq(snd_ctl_elem_value_get_enumerated(elem, 0),
+			sco_codec_enum_items > 1 ? 1 : 0);
 #if ENABLE_MSBC
 	/* select SCO CVSD codec */
 	snd_ctl_elem_value_set_enumerated(elem, 0, 0);
@@ -440,9 +465,9 @@ CK_START_TEST(test_single_device_non_dynamic) {
 	ck_assert_int_eq(snd_ctl_elem_value_get_integer(elem_volume, 0), 42);
 
 	/* Process events until we will be notified about A2DP profile disconnection.
-	 * We shall get 2 events from previous value update and 2 events for profile
+	 * We shall get 4 events from previous value update and 2 events for profile
 	 * disconnection (one event per switch/volume element). */
-	for (size_t events = 0; events < 4;) {
+	for (size_t events = 0; events < 4 + 2 + 2;) {
 		ck_assert_int_eq(snd_ctl_wait(ctl, 750), 1);
 		while (snd_ctl_read(ctl, event) == 1)
 			events++;
@@ -479,7 +504,7 @@ CK_START_TEST(test_notifications) {
 				"--fuzzing=250",
 				NULL), -1);
 
-	ck_assert_int_eq(snd_ctl_open(&ctl, "bluealsa:BAT=yes", 0), 0);
+	ck_assert_int_eq(snd_ctl_open(&ctl, "bluealsa:EXT=battery", 0), 0);
 
 	snd_ctl_event_t *event;
 	snd_ctl_event_malloc(&event);
@@ -490,14 +515,22 @@ CK_START_TEST(test_notifications) {
 	while (snd_ctl_wait(ctl, 500) == 1)
 		while (snd_ctl_read(ctl, event) == 1) {
 			ck_assert_int_eq(snd_ctl_event_get_type(event), SND_CTL_EVENT_ELEM);
+			debug("Event: %s: %s", test_ctl_event_elem_get_mask_name(event),
+					snd_ctl_event_elem_get_name(event));
 			events++;
 		}
 
 	ck_assert_int_eq(snd_ctl_subscribe_events(ctl, 0), 0);
 
 	size_t events_update_codec = 0;
-#if ENABLE_MSBC
-	events_update_codec += 8;
+#if ENABLE_HFP_CODEC_SELECTION
+	events_update_codec += 4;
+# if ENABLE_MSBC
+	events_update_codec += 4;
+# endif
+# if ENABLE_LC3_SWB
+	events_update_codec += 4;
+# endif
 #endif
 
 	/* Processed events:
@@ -505,18 +538,20 @@ CK_START_TEST(test_notifications) {
 	 * - 2 removes; 4 new elems (12:34:... A2DP, 23:45:... A2DP)
 	 * - 4 removes; 7 new elems (2x A2DP, SCO playback, battery)
 	 * - 7 removes; 9 new elems (2x A2DP, SCO playback/capture, battery)
-	 * - 8 updates (SCO codec updates if mSBC is supported)
-	 *
-	 * XXX: It is possible that the battery element (RFCOMM D-Bus path) will not
+	 * - 4 updates per codec (SCO codec updates if codec selection is supported)
+	 */
+	size_t expected_events = (0 + 2) + (2 + 4) + (4 + 7) + (7 + 9) + events_update_codec;
+
+	/* XXX: It is possible that the battery element (RFCOMM D-Bus path) will not
 	 *      be exported in time. In such case, the number of events will be less
 	 *      by 2 when RFCOMM D-Bus path is not available during the playback SCO
 	 *      addition and less by another 1 when the path is not available during
 	 *      the capture SCO addition. We shall account for this in the test, as
 	 *      it is not an error. */
-	if (events == (35 + events_update_codec - 2) ||
-			events == (35 + events_update_codec - 2 - 1))
-		events = 35 + events_update_codec;
-	ck_assert_int_eq(events, (0 + 2) + (2 + 4) + (4 + 7) + (7 + 9) + events_update_codec);
+	int result = events == expected_events ||
+					events == expected_events - 2 ||
+					events == expected_events - 3;
+	ck_assert_int_eq(result, 1);
 
 	snd_ctl_event_free(event);
 	ck_assert_int_eq(test_pcm_close(&sp_ba_mock, ctl), 0);
@@ -541,8 +576,8 @@ CK_START_TEST(test_alsa_high_level_control_interface) {
 
 } CK_END_TEST
 
-int main(int argc, char *argv[], char *envp[]) {
-	preload(argc, argv, envp, ".libs/aloader.so");
+int main(int argc, char *argv[]) {
+	preload(argc, argv, ".libs/libaloader.so");
 
 	char *argv_0 = strdup(argv[0]);
 	snprintf(bluealsa_mock_path, sizeof(bluealsa_mock_path),
@@ -554,6 +589,10 @@ int main(int argc, char *argv[], char *envp[]) {
 
 	suite_add_tcase(s, tc);
 
+	/* Boost timeout since the single_device_non_dynamic test takes more than
+	 * 3.5 seconds and from time to time can exceed default 4 seconds. */
+	tcase_set_timeout(tc, 6);
+
 	tcase_add_test(tc, test_controls);
 	tcase_add_test(tc, test_controls_battery);
 	tcase_add_test(tc, test_controls_extended);
