@@ -1,6 +1,6 @@
 /*
  * BlueALSA - alsa-pcm.h
- * Copyright (c) 2016-2020 Arkadiusz Bokowy
+ * Copyright (c) 2016-2023 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -8,9 +8,12 @@
  *
  */
 
+#pragma once
 #ifndef BLUEALSA_APLAY_ALSAPCM_H_
 #define BLUEALSA_APLAY_ALSAPCM_H_
 
+#include <stdio.h>
+
 #include <alsa/asoundlib.h>
 
 int alsa_pcm_open(snd_pcm_t **pcm, const char *name,
@@ -18,4 +21,6 @@ int alsa_pcm_open(snd_pcm_t **pcm, const char *name,
 		unsigned int *buffer_time, unsigned int *period_time,
 		char **msg);
 
+void alsa_pcm_dump(snd_pcm_t *pcm, FILE *fp);
+
 #endif
